###############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

import datetime
import json

from girder import events, logger
from girder.constants import AccessType
from girder.plugin import GirderPlugin, getPlugin
from girder_jobs.constants import JobStatus
from girder_jobs.models.job import Job

from . import worker_tools
from .docker_resource import DockerResource
from .models import DockerImageItem


def _onUpload(event):
    try:
        ref = json.loads(event.info.get('reference'))
    except (ValueError, TypeError):
        return

    if isinstance(ref, dict) and ref.get('type') == 'slicer_cli.parameteroutput':
        job = Job().load(ref['jobId'], force=True, exc=True)

        file = event.info['file']

        # Add link to job model to the output item
        Job().updateJob(job, otherFields={
            'slicerCLIBindings.outputs.parameters': file['_id']
        })


class SlicerCLIWebPlugin(GirderPlugin):
    DISPLAY_NAME = 'Slicer CLI Web'
    CLIENT_SOURCE_PATH = 'web_client'

    def load(self, info):
        try:
            getPlugin('worker').load(info)
        except Exception:
            logger.info('Girder working is unavailable')

        DockerImageItem.prepare()

        # resource name must match the attribute added to info[apiroot]
        resource = DockerResource('slicer_cli_web')
        info['apiRoot'].slicer_cli_web = resource

        Job().exposeFields(level=AccessType.READ, fields={'slicerCLIBindings'})

        events.bind('jobs.job.update.after', resource.resourceName,
                    resource.addRestEndpoints)
        events.bind('data.process', 'slicer_cli_web', _onUpload)

        count = 0
        for job in Job().find({
            'status': {'$in': [
                JobStatus.INACTIVE, JobStatus.QUEUED, JobStatus.RUNNING,
                # from girder_worker, but we don't strictly require its
                # existence
                820, 821, 822, 823, 824,
            ]},
            'updated': {'$lt': datetime.datetime.utcnow() - datetime.timedelta(days=7)}
        }, force=True):
            try:
                Job().updateJob(job, log='Canceled stale job.', status=JobStatus.CANCELED)
                count += 1
            except Exception:
                pass
        if count:
            logger.info('Marking %d old job(s) as cancelled' % count)
        worker_tools.start()



#############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#############################################################################


import json
import os
import re

import pymongo
from girder.api import access
from girder.api.describe import Description, autoDescribeRoute, describeRoute
from girder.api.rest import setRawResponse, setResponseHeader
from girder.api.v1.resource import Resource, RestException
from girder.constants import AccessType, SortDir
from girder.exceptions import AccessException
from girder.models.item import Item
from girder.utility import path as path_util
from girder.utility.model_importer import ModelImporter
from girder_jobs.constants import JobStatus
from girder_jobs.models.job import Job

from .config import PluginSettings
from .models import CLIItem, DockerImageItem, DockerImageNotFoundError
from .rest_slicer_cli import genRESTEndPointsForSlicerCLIsForItem


class DockerResource(Resource):
    """
    Resource object that handles runtime generation and deletion of rest
    endpoints
    """

    jobType = 'slicer_cli_web_job'

    def __init__(self, name):
        super().__init__()
        self.currentEndpoints = {}
        self.resourceName = name
        self.jobType = 'slicer_cli_web_job'
        self.route('PUT', ('docker_image',), self.setImages)
        self.route('DELETE', ('docker_image',), self.deleteImage)
        self.route('GET', ('docker_image',), self.getDockerImages)

        self.route('GET', ('cli',), self.getItems)
        self.route('GET', ('cli', ':id',), self.getItem)
        self.route('DELETE', ('cli', ':id',), self.deleteItem)
        # run is generated per item for better validation
        self.route('GET', ('cli', ':id', 'xml'), self.getItemXML)

        self.route('GET', ('path_match', ), self.getMatchingResource)

        self._generateAllItemEndPoints()

    @access.public
    @describeRoute(
        Description('List docker images and their CLIs')
        .notes('You must be logged in to see any results.')
    )
    def getDockerImages(self, params):
        data = {}
        if self.getCurrentUser():
            for image in DockerImageItem.findAllImages(self.getCurrentUser()):
                imgData = {}
                for cli in image.getCLIs():
                    basePath = '/%s/cli/%s' % (self.resourceName, cli._id)
                    imgData[cli.name] = {
                        'type': cli.type,
                        'xmlspec': basePath + '/xml',
                        'run': basePath + '/run'
                    }
                data.setdefault(image.image, {})[image.tag] = imgData
        return data

    @access.admin
    @describeRoute(
        Description('Remove a docker image')
        .notes('Must be a system administrator to call this.')
        .param('name', 'The name or a list of names of the docker images to be '
               'removed', required=True)
        .param('delete_from_local_repo',
               'If True the image is deleted from the local repo, requiring '
               'it to be pulled from a remote repository the next time it is '
               'used.  If False the metadata regarding the image is deleted, '
               'but the docker image remains.', required=False,
               dataType='boolean', default=False)
        .errorResponse('You are not a system administrator.', 403)
        .errorResponse('Failed to set system setting.', 500)
    )
    def deleteImage(self, params):
        self.requireParams(('name',), params)
        if 'delete_from_local_repo' in params:
            deleteImage = str(params['delete_from_local_repo']).lower() == 'true'
        else:
            deleteImage = False
        nameList = self.parseImageNameList(params['name'])
        self._deleteImage(nameList, deleteImage)

    def _deleteImage(self, names, deleteImage):
        """
        Removes the docker images and there respective clis endpoints

        :param names: The name of the docker image (user/rep:tag)
        :type name: list
        :param deleteImage: Boolean indicating whether to delete the docker
            image from the local machine.(if True this is equivalent to
            docker rmi -f <image> )
        :type name: boolean
        """
        removed = DockerImageItem.removeImages(names, self.getCurrentUser())
        if removed != names:
            rest = [name for name in names if name not in removed]
            raise RestException('Some docker images could not be removed. %s' % (rest))
        self.deleteImageEndpoints(removed)

        try:
            if deleteImage:
                self._deleteDockerImages(removed)
        except DockerImageNotFoundError as err:
            raise RestException('Invalid docker image name. ' + str(err))

    def _deleteDockerImages(self, removed):
        """
        Creates an asynchronous job to delete the docker images listed in name
        from the local machine
        :param removed:A list of docker image names
        :type removed: list of strings
        """
        job = Job().createLocalJob(
            module='slicer_cli_web.image_job',

            function='deleteImage',
            kwargs={
                'deleteList': removed
            },
            title='Deleting Docker Images',
            user=self.getCurrentUser(),
            type=self.jobType,
            public=True,
            asynchronous=True
        )

        Job().scheduleJob(job)

    def parseImageNameList(self, param):
        """
        Parse a string to get a list of image names.  If the string is a JSON
        list of strings or a JSON string (with quotes), it is processed as
        JSON.  Otherwise, the input value is treated as a single image name.

        :param param: a parameter with an image name, a JSON image name, or a
            JSON list of image names.
        :returns: a list of image names.
        """
        nameList = param
        if isinstance(param, bytes):
            param = param.decode('utf8')
        if isinstance(param, str):
            try:
                nameList = json.loads(param)
            except ValueError:
                pass
        if isinstance(nameList, str):
            nameList = [nameList]
        if not isinstance(nameList, list):
            raise RestException('A valid string or a list of strings is required.')
        for img in nameList:
            if not isinstance(img, str):
                raise RestException('%r is not a valid string.' % img)
            if ':' not in img and '@' not in img:
                raise RestException('Image %s does not have a tag or digest' % img)
        return nameList

    @access.admin
    @describeRoute(
        Description('Add one or a list of images')
        .notes('Must be a system administrator to call this.')
        .param('name', 'A name or a list of names of the docker images to be '
               'loaded', required=True)
        .modelParam('folder', 'The base folder id to upload the tasks to',
                    'folder', paramType='query',
                    level=AccessType.WRITE, required=False)
        .param('pull', 'If True, try to repull all images', paramType='query',
               required=False)
        .errorResponse('You are not a system administrator.', 403)
        .errorResponse('Failed to set system setting.', 500)
    )
    def setImages(self, params):
        """
        Validates the new images to be added (if they exist or not) and then
        attempts to collect xml data to be cached. a job is then called to
        update the girder collection containing the cached docker image data
        """
        self.requireParams(('name',), params)
        nameList = self.parseImageNameList(params['name'])
        folder = params.get('folder', PluginSettings.get_task_folder())
        if not folder:
            raise RestException('no upload folder given or defined by default')
        return self._createPutImageJob(nameList, folder, params.get('pull', None))

    def _createPutImageJob(self, nameList, baseFolder, pull=False):
        job = Job().createLocalJob(
            module='slicer_cli_web.image_job',
            function='jobPullAndLoad',
            kwargs={
                'nameList': nameList,
                'folder': baseFolder['_id'] if isinstance(baseFolder, dict) else baseFolder,
                'pull': pull,
            },
            title='Pulling and caching docker images',
            type=self.jobType,
            user=self.getCurrentUser(),
            public=True,
            asynchronous=True
        )
        Job().scheduleJob(job)
        return job

    def storeEndpoints(self, imgName, cliName, undoFunction):
        """
        Information on each rest endpoint is saved so they can be
        deleted when docker images are removed or loaded.

        :param imgName: The full name of the docker image with the tag.
            This name must match exactly with the name the command
            docker images displays in the console
        :type imgName: string
        :param cliName: The name of the cli whose rest endpoint is being stored. The
            cli must match exactly with what the docker image returns when
            running <docker image> --list_cli
        :type cliName: string
        """
        img = self.currentEndpoints.setdefault(imgName, {})
        img[cliName] = undoFunction

    def deleteImageEndpoints(self, imageList=None):

        if imageList is None:
            imageList = self.currentEndpoints.keys()
        for imageName in list(imageList):
            for undoFunction in self.currentEndpoints.pop(imageName, {}).values():
                undoFunction()

    def _generateAllItemEndPoints(self):
        # sort by name and creation date desc
        items = sorted(CLIItem.findAllItems(), key=lambda x: (x.restPath, x.item['created']),
                       reverse=True)

        seen = set()
        for item in items:
            # default if not seen yet
            genRESTEndPointsForSlicerCLIsForItem(self, item, item.restPath not in seen)
            seen.add(item.restPath)

    def addRestEndpoints(self, event):
        """
        Determines if the job event being triggered is due to the caching of
        new docker images or deleting a docker image off the local machine.  If
        a new image is being loaded all old rest endpoints are deleted and
        endpoints for all cached docker images are regenerated.

        :param event: An event dictionary
        """
        job = event.info['job']

        if job['type'] == self.jobType and job['status'] == JobStatus.SUCCESS:
            self.deleteImageEndpoints()
            self._generateAllItemEndPoints()

    def _dump(self, item, details=False):
        r = {
            '_id': item._id,
            'name': item.name,
            'type': item.type,
            'image': item.image,
            'description': item.item['description']
        }
        if details:
            r['xml'] = item.item['meta']['xml']
        return r

    @access.user
    @autoDescribeRoute(
        Description('List CLIs')
        .errorResponse('You are not logged in.', 403)
        .modelParam('folder', 'The base folder to look for tasks', 'folder', paramType='query',
                    level=AccessType.WRITE, required=False)
    )
    def getItems(self, folder):
        items = CLIItem.findAllItems(self.getCurrentUser(), baseFolder=folder)
        return [self._dump(item) for item in items]

    @access.user
    @autoDescribeRoute(
        Description('Get a specific CLI')
        .errorResponse('You are not logged in.', 403)
        .modelParam('id', 'The task item', 'item',
                    level=AccessType.READ)
    )
    def getItem(self, item):
        return self._dump(CLIItem(item), True)

    @access.user
    @autoDescribeRoute(
        Description('Get a specific CLI')
        .errorResponse('You are not logged in.', 403)
        .modelParam('id', 'The task item', 'item',
                    level=AccessType.WRITE)
    )
    def deleteItem(self, item):
        Item().remove(item)
        return dict(status='OK')

    @access.user
    @autoDescribeRoute(
        Description('Get a specific CLI')
        .errorResponse('You are not logged in.', 403)
        .modelParam('id', 'The task item', 'item',
                    level=AccessType.READ)
    )
    def getItemXML(self, item):
        setResponseHeader('Content-Type', 'application/xml')
        setRawResponse()
        return CLIItem(item).xml

    @access.public
    @autoDescribeRoute(
        Description(
            'Get the most recently updated resource that has a name and path '
            'that matches a regular expression')
        .notes('This can be very slow if name is too general.')
        .param('name', 'A regular expression to match the name of the '
               'resource.', required=False)
        .param('path', 'A regular expression to match the entire resource path.', required=False)
        .param('relative_path', 'A relative resource path to the base item.', required=False)
        .param('base_id', 'The base girder id for the relative path', required=False)
        .param('base_type', 'The base girder type for the relative path', required=False)
        .param('type', 'The type of the resource (item, file, etc.).')
        .errorResponse('Invalid resource type.')
        .errorResponse('No matches.')
    )
    def getMatchingResource(self, name, path, type, relative_path, base_id, base_type):
        user = self.getCurrentUser()
        model = ModelImporter.model(type)
        pattern = None
        if path:
            pattern = re.compile(path)
        if relative_path:
            if not base_id or not base_type:
                return None
            try:
                base_model = ModelImporter.model(base_type).load(base_id, user=user)
                base_path = path_util.getResourcePath(base_type, base_model, user=user)
                new_path = os.path.normpath(os.path.join(base_path, relative_path))
                doc = path_util.lookUpPath(new_path, user=user)['document']
                doc['_path'] = new_path.split('/')[2:]
                if type == 'folder':
                    doc['_path'] = doc['_path'][:-1]
            except Exception:
                return None
            if not name and not path:
                return doc
            pattern = re.compile('(?=^' + re.escape(new_path) + ').*' + (path or ''))
        try:
            for doc in model.findWithPermissions(
                    {'name': {'$regex': name}} if name else {},
                    sort=[('updated', SortDir.DESCENDING), ('created', SortDir.DESCENDING)],
                    user=user, level=AccessType.READ, timeout=10000):
                try:
                    resourcePath = path_util.getResourcePath(type, doc, user=user)
                    if not pattern or pattern.search(resourcePath):
                        doc['_path'] = resourcePath.split('/')[2:]
                        if type == 'folder':
                            doc['_path'] = doc['_path'][:-1]
                        return doc
                except (AccessException, TypeError):
                    pass
        except pymongo.errors.ExecutionTimeout:
            return None
        return None



import functools
import json
import os
import time

import pytest
from girder.api import rest
from girder.models.collection import Collection
from girder.models.folder import Folder
from girder.models.item import Item
from pytest_girder.assertions import assertStatusOk

from slicer_cli_web import docker_resource, rest_slicer_cli
from slicer_cli_web.models import CLIItem


@pytest.fixture
def handlerFunc(server, admin, folder, file):
    # Make a request to allow there to be some lingering context to handle the
    # testing outside of a request for the actual handler.
    server.request('/system/version')
    rest.setCurrentUser(admin)

    xmlpath = os.path.join(os.path.dirname(__file__), 'data', 'ExampleSpec.xml')

    girderCLIItem = Item().createItem('data', admin, folder)
    Item().setMetadata(girderCLIItem, dict(
        slicerCLIType='task', type='python', image='dockerImage',
        digest='dockerImage@sha256:abc', xml=open(xmlpath, 'rb').read()))

    resource = docker_resource.DockerResource('test')
    item = CLIItem(girderCLIItem)
    handlerFunc = rest_slicer_cli.genHandlerToRunDockerCLI(item)
    yield functools.partial(handlerFunc, resource)


@pytest.fixture
def handlerRerunFuncs(server, admin, folder, file):
    # Make a request to allow there to be some lingering context to handle the
    # testing outside of a request for the actual handler.
    server.request('/system/version')
    rest.setCurrentUser(admin)

    xmlpath = os.path.join(os.path.dirname(__file__), 'data', 'ExampleSpec.xml')

    girderCLIItem = Item().createItem('data', admin, folder)
    Item().setMetadata(girderCLIItem, dict(
        slicerCLIType='task', type='python', image='dockerImage',
        digest='dockerImage@sha256:abc', xml=open(xmlpath, 'rb').read()))

    resource = docker_resource.DockerResource('test')
    item = CLIItem(girderCLIItem)
    handlerFunc = rest_slicer_cli.genHandlerToRunDockerCLI(item)
    handlerRerunFunc = rest_slicer_cli.genHandlerToReRunDockerCLI(item, handlerFunc)
    yield functools.partial(handlerFunc, resource), functools.partial(handlerRerunFunc, resource)


@pytest.mark.plugin('slicer_cli_web')
def test_genHandlerToRunDockerCLI(folder, file, handlerFunc):
    assert handlerFunc is not None

    job = handlerFunc(params={
        'inputImageFile': str(file['_id']),
        'secondImageFile': str(file['_id']),
        'outputStainImageFile_1_folder': str(folder['_id']),
        'outputStainImageFile_1': 'sample1.png',
        'outputStainImageFile_2_folder': str(folder['_id']),
        'outputStainImageFile_2_name': 'sample2.png',
        'stainColor_1': '[0.5, 0.5, 0.5]',
        'stainColor_2': '[0.2, 0.3, 0.4]',
        'returnparameterfile_folder': str(folder['_id']),
        'returnparameterfile': 'output.data',
    })

    kwargs = json.loads(job['kwargs'])
    assert 'container_args' in kwargs
    assert 'image' in kwargs
    assert 'pull_image' in kwargs

    assert kwargs['image'] == 'dockerImage@sha256:abc'
    assert kwargs['pull_image'] == 'if-not-present'
    container_args = kwargs['container_args']
    assert container_args[0] == 'data'


@pytest.mark.plugin('slicer_cli_web')
def test_get_matching_resource(server, admin):
    # Make some resources
    publicFolder = Folder().find({'parentId': admin['_id'], 'name': 'Public'})[0]
    privateFolder = Folder().find({'parentId': admin['_id'], 'name': 'Private'})[0]
    # Create some resources to use in the tests; this was mostly copied from a
    # test in another repo and has more items and folders than needed
    collection = Collection().createCollection(
        'collection A', admin)
    colFolderA = Folder().createFolder(
        collection, 'folder A', parentType='collection',
        creator=admin)
    colFolderB = Folder().createFolder(
        collection, 'folder B', parentType='collection',
        creator=admin)
    colFolderC = Folder().createFolder(
        colFolderA, 'folder C', creator=admin)
    Item().createItem('item A1', admin, colFolderA)
    Item().createItem('item B1', admin, colFolderB)
    Item().createItem('item B2', admin, colFolderB)
    Item().createItem('item C1', admin, colFolderC)
    Item().createItem('item C2', admin, colFolderC)
    Item().createItem('item C3', admin, colFolderC)
    Item().createItem('item Public 1', admin, publicFolder)
    Item().createItem('item Private 1', admin, privateFolder)
    folderD = Folder().createFolder(publicFolder, 'folder D', creator=admin)
    Item().createItem('item D1', admin, folderD)
    itemD2 = Item().createItem('item D2', admin, folderD)
    # Now test
    resp = server.request('/slicer_cli_web/path_match', params={
        'type': 'folder',
        'name': '^fold.*D',
    }, user=admin)
    assertStatusOk(resp)
    assert resp.json['name'] == 'folder D'

    resp = server.request('/slicer_cli_web/path_match', params={
        'type': 'item',
        'name': '^item',
        'path': '^/collection/.*/folder.*/folder',
    }, user=admin)
    assertStatusOk(resp)
    assert resp.json['name'] == 'item C3'

    resp = server.request('/slicer_cli_web/path_match', params={
        'type': 'item',
        'name': '^item .* 1',
    }, user=admin)
    assertStatusOk(resp)
    assert resp.json['name'] == 'item Private 1'

    resp = server.request('/slicer_cli_web/path_match', params={
        'type': 'item',
        'name': '^item .* 1',
    })
    assertStatusOk(resp)
    assert resp.json['name'] == 'item Public 1'

    resp = server.request('/slicer_cli_web/path_match', params={
        'type': 'item',
        'name': 'nosuchitem',
    })
    assertStatusOk(resp)
    assert resp.json is None

    resp = server.request('/slicer_cli_web/path_match', params={
        'type': 'folder',
        'relative_path': '..',
        'base_type': 'item',
        'base_id': itemD2['_id']
    }, user=admin)
    assertStatusOk(resp)
    assert resp.json['name'] == 'folder D'

    resp = server.request('/slicer_cli_web/path_match', params={
        'type': 'folder',
        'relative_path': '..',
        'base_type': 'item',
    }, user=admin)
    assertStatusOk(resp)
    assert resp.json is None

    resp = server.request('/slicer_cli_web/path_match', params={
        'type': 'folder',
        'relative_path': '..',
        'base_type': 'item',
        'base_id': 'nosuchid'
    }, user=admin)
    assertStatusOk(resp)
    assert resp.json is None


@pytest.mark.plugin('slicer_cli_web')
def test_girderApiUrl(handlerFunc, folder, file):
    job = handlerFunc(params={
        'inputImageFile': str(file['_id']),
        'secondImageFile': str(file['_id']),
        'outputStainImageFile_1_folder': str(folder['_id']),
        'outputStainImageFile_1': 'sample1.png',
        'outputStainImageFile_2_folder': str(folder['_id']),
        'outputStainImageFile_2_name': 'sample2.png',
        'stainColor_1': '[0.5, 0.5, 0.5]',
        'stainColor_2': '[0.2, 0.3, 0.4]',
        'returnparameterfile_folder': str(folder['_id']),
        'returnparameterfile': 'output.data',
        'girderApiUrl': '',
        'girderToken': '',
    })
    kwargs = json.loads(job['kwargs'])
    assert '--api-url' in kwargs['container_args']
    assert '--girder-token' in kwargs['container_args']


@pytest.mark.parametrize('testParams,results', [
    ({'outputStainImageFile_1': '{{reference_base}}{{extension}}'}, 'Sample.jpg'),
    ({'outputStainImageFile_1': '{{task}}.test.png'}, 'data.test.png'),
    ({'outputStainImageFile_1': '{{title}}.test.png'},
     'Performs Adaptive Color Deconvolution.test.png'),
    ({'outputStainImageFile_1': '{{image}}.test.png'}, 'dockerImage.test.png'),
    ({'outputStainImageFile_1': 'test.{{now}}.png'}, 'test.' + time.strftime('%Y')),
    ({'outputStainImageFile_1': '{{yyyy}}.test.png'}, time.strftime('%Y') + '.test.png'),
    ({'outputStainImageFile_1': '{{parameter_stainColor_1}}.png'}, '0.5, 0.5, 0.5.png'),
    ({'outputStainImageFile_1': '{{parameter_inputImageFile}}.png'}, 'Sample.png'),
    ({'outputStainImageFile_1': '{{parameter_inputImageFile_base}}.png'}, 'Sample.png'),
    ({'outputStainImageFile_1': '{{name}}.png'}, 'outputStainImageFile_1.png'),
    ({'outputStainImageFile_1': '{{label}}.png'}, 'Output image of Stain 1.png'),
    ({'outputStainImageFile_1': '{{description}}.png'}, 'Output Image of Stain 1.png'),
    ({'outputStainImageFile_1': '{{index}}.png'}, '2.png'),
    ({'gamma': '{{default}}0'}, '0.50'),
    ({'gamma': '{{env_GAMMA}}'}, '0.4'),
    ({'gamma': '{{env_NOTGAMMA|default("0.3")}}'}, '0.3'),
    ({'gamma': '{#control:#some_control#}'}, ''),
])
@pytest.mark.plugin('slicer_cli_web')
def test_templateParams(handlerFunc, folder, file, testParams, results):
    os.environ['SLICER_CLI_WEB_GAMMA'] = '0.4'
    job = handlerFunc(params=dict({
        'inputImageFile': str(file['_id']),
        'secondImageFile': str(file['_id']),
        'outputStainImageFile_1_folder': str(folder['_id']),
        'outputStainImageFile_1': 'sample1.png',
        'outputStainImageFile_2_folder': str(folder['_id']),
        'outputStainImageFile_2_name': 'sample2.png',
        'stainColor_1': '[0.5, 0.5, 0.5]',
        'stainColor_2': '[0.2, 0.3, 0.4]',
        'returnparameterfile_folder': str(folder['_id']),
        'returnparameterfile': 'output.data',
    }, **testParams))
    ca = json.loads(job['kwargs'])['container_args']
    assert results in repr(ca)


@pytest.mark.plugin('slicer_cli_web')
def test_genHandlerToReRunDockerCLI(folder, file, handlerRerunFuncs):
    handlerFunc, handlerRerunFunc = handlerRerunFuncs
    assert handlerFunc is not None
    assert handlerRerunFunc is not None

    job = handlerFunc(params={
        'inputImageFile': str(file['_id']),
        'secondImageFile': str(file['_id']),
        'outputStainImageFile_1_folder': str(folder['_id']),
        'outputStainImageFile_1': 'sample1.png',
        'outputStainImageFile_2_folder': str(folder['_id']),
        'outputStainImageFile_2_name': 'sample2.png',
        'stainColor_1': '[0.5, 0.5, 0.5]',
        'stainColor_2': '[0.2, 0.3, 0.4]',
        'returnparameterfile_folder': str(folder['_id']),
        'returnparameterfile': 'output.data',
    })

    rerunJob = handlerRerunFunc(params={'jobId': str(job['_id'])})

    kwargs = json.loads(rerunJob['kwargs'])
    assert 'container_args' in kwargs
    assert 'image' in kwargs
    assert 'pull_image' in kwargs

    assert kwargs['image'] == 'dockerImage@sha256:abc'
    assert kwargs['pull_image'] == 'if-not-present'
    container_args = kwargs['container_args']
    assert container_args[0] == 'data'



import io
import json
import os
import subprocess
import threading
import time
import types

import pytest
from girder import events
from girder.models.folder import Folder
from girder.models.item import Item
from girder.models.setting import Setting
from girder.models.token import Token
from girder.models.upload import Upload
from girder_jobs.constants import JobStatus
from girder_jobs.models.job import Job
from girder_worker.girder_plugin.constants import PluginSettings as WorkerSettings
from pytest_girder.assertions import assertStatus, assertStatusOk
from pytest_girder.utils import getResponseBody


@pytest.fixture
def folder(admin):
    f = Folder().createFolder(admin, 'folder', parentType='user')

    yield f


@pytest.fixture
def item(folder, admin):
    f = Item().createItem('item', admin, folder)

    yield f


@pytest.fixture
def file(folder, admin, fsAssetstore):

    sampleData = b'Hello world'
    f = Upload().uploadFromFile(
        obj=io.BytesIO(sampleData), size=len(sampleData), name='Sample',
        parentType='folder', parent=folder, user=admin)

    yield f


@pytest.fixture
def adminToken(admin):
    yield Token().createToken(admin)


def splitName(name):
    if ':' in name.split('/')[-1]:
        imageAndTag = name.rsplit(':', 1)
    else:
        imageAndTag = name.split('@')
    return imageAndTag[0], imageAndTag[1]


class ImageHelper:
    TIMEOUT = 180

    def __init__(self, server, admin, folder):
        self.server = server
        self.admin = admin
        self.folder = folder

    def assertHasKeys(self, obj, keys):
        for key in keys:
            assert key in obj

    def assertNotHasKeys(self, obj, keys):
        for key in keys:
            assert key not in obj

    def getEndpoint(self):
        resp = self.server.request(path='/slicer_cli_web/docker_image',
                                   user=self.admin)
        assertStatus(resp, 200)

        return json.loads(getResponseBody(resp))

    def assertNoImages(self):
        data = self.getEndpoint()
        assert {} == data, 'There should be no pre existing docker images'

    def imageIsLoaded(self, name, exists):
        userAndRepo, tag = splitName(name)

        data = self.getEndpoint()
        if not exists:
            if userAndRepo in data:
                imgVersions = data[userAndRepo]
                self.assertNotHasKeys(imgVersions, [tag])
        else:
            self.assertHasKeys(data, [userAndRepo])
            imgVersions = data[userAndRepo]
            self.assertHasKeys(imgVersions, [tag])

    def endpointsExist(self, name, present=None, absent=None):
        """
        Test if endpoints for particular image exist.

        :param name: name of the image used to determine endpoint location.
        :param present: a list of endpoints within the image that must exist.
        :param absent: a list of endpoints that should be in the image but not
            have endpoints.
        """
        present = present or []
        absent = absent or []
        userAndRepo, tag = splitName(name)
        data = self.getEndpoint()
        for cli in present:
            self.assertHasKeys(data, [userAndRepo])
            self.assertHasKeys(data[userAndRepo], [tag])
            self.assertHasKeys(data[userAndRepo][tag], [cli])
            path = data[userAndRepo][tag][cli]['xmlspec']
            resp = self.server.request(path=path, user=self.admin, isJson=False)
            assertStatusOk(resp)
        for cli in absent:
            self.assertHasKeys(data, [userAndRepo])
            self.assertHasKeys(data[userAndRepo], [tag])
            self.assertNotHasKeys(data[userAndRepo][tag], [cli])

    def deleteImage(self, name, responseCodeOK, deleteDockerImage=False,
                    status=4):
        """
        Delete docker image data and test whether a docker
        image can be deleted off the local machine
        """
        job_status = [JobStatus.SUCCESS]
        if deleteDockerImage:
            event = threading.Event()

            def tempListener(self, girderEvent):
                job = girderEvent.info['job']

                if (job['type'] == 'slicer_cli_web_job' and
                        job['status'] in (JobStatus.SUCCESS, JobStatus.ERROR)):
                    assert job['status'] == status, 'The status of the job should match'
                    events.unbind('jobs.job.update.after', 'slicer_cli_web_del')
                    job_status[0] = job['status']
                    event.set()

            self.delHandler = types.MethodType(tempListener, self)

            events.bind('jobs.job.update.after', 'slicer_cli_web_del',
                        self.delHandler)

        resp = self.server.request(
            path='/slicer_cli_web/docker_image', user=self.admin,
            method='DELETE', params={
                'name': json.dumps(name),
                'delete_from_local_repo': deleteDockerImage,
            }, isJson=False)
        if responseCodeOK:
            assertStatus(resp, 200)
        else:
            assertStatus(resp, 400)
            # A status ok or code 200 should not have been received for
            # deleting the image %s' % str(name))

        if deleteDockerImage:
            if not event.wait(self.TIMEOUT):
                del self.delHandler
                raise AssertionError(
                    'deleting the docker image is taking longer than %d seconds' % self.TIMEOUT)
            else:
                del self.delHandler
                assert job_status[0] == status, 'The status of the job should match '

    def addImage(self, name, status, initialStatus=200):
        """
        Test the put endpoint.

        :param name: a string or a list of strings
        :param status: either JobStatus.SUCCESS or JobStatus.ERROR.
        :param initialStatus: 200 if the job should run, otherwise a HTTP error
            code expected if the job will fail.
        """
        event = threading.Event()
        job_status = [JobStatus.SUCCESS]

        def tempListener(self, girderEvent):
            job = girderEvent.info['job']

            if (job['type'] == 'slicer_cli_web_job' and
                    job['status'] in (JobStatus.SUCCESS, JobStatus.ERROR)):
                assert job['status'] == status, 'The status of the job should match'
                job_status[0] = job['status']

                events.unbind('jobs.job.update.after', 'slicer_cli_web_add')

                # wait 10sec before continue
                threading.Timer(5, lambda: event.set()).start()

        if initialStatus == 200:
            self.addHandler = types.MethodType(tempListener, self)

            events.bind('jobs.job.update.after',
                        'slicer_cli_web_add', self.addHandler)

        resp = self.server.request(
            path='/slicer_cli_web/docker_image',
            user=self.admin, method='PUT', params={'name': json.dumps(name),
                                                   'folder': self.folder['_id']},
            isJson=initialStatus == 200)

        assertStatus(resp, initialStatus)
        if initialStatus != 200:
            return
        # We should have a job ID
        assert resp.json.get('_id') is not None

        if not event.wait(self.TIMEOUT):
            del self.addHandler
            raise AssertionError(
                'adding the docker image is taking longer than %d seconds' % self.TIMEOUT)
        else:
            del self.addHandler
            assert job_status[0] == status, 'The status of the job should match '


@pytest.fixture
def images(server, admin, folder):
    return ImageHelper(server, admin, folder)


@pytest.fixture
def fileset(db, admin, fsAssetstore):
    f1 = Folder().createFolder(admin, 'folder1', parentType='user')
    f2 = Folder().createFolder(admin, 'folder2', parentType='user')
    results = {
        'folder1': f1,
        'folder2': f2,
        'item1': [],
        'item2': [],
        'file1': [],
        'file2': [],
    }
    for fidx, folder in enumerate([f1, f2]):
        for idx in range(10):
            sampleData = b'Sample %d' % idx
            f = Upload().uploadFromFile(
                obj=io.BytesIO(sampleData), size=len(sampleData),
                name='Sample %d' % idx, parentType='folder', parent=folder,
                user=admin)
            results['file%d' % (fidx + 1)].append(f)
            i = Item().load(f['itemId'], force=True)
            if not (idx % 2):
                i['largeImage'] = {'fileId': str(f['_id'])}
                i = Item().save(i)
            results['item%d' % (fidx + 1)].append(i)
    yield results


@pytest.fixture
def girderWorker(db):
    """
    Run an instance of Girder worker, connected to rabbitmq.  The rabbitmq
    service must be running.
    """
    broker = 'amqp://guest@127.0.0.1'
    backend = 'rpc://guest@127.0.0.1'
    Setting().set(WorkerSettings.BROKER, broker)
    Setting().set(WorkerSettings.BACKEND, backend)
    env = os.environ.copy()
    env['C_FORCE_ROOT'] = 'true'
    proc = subprocess.Popen([
        'celery', '-A', 'girder_worker.app', '--broker', broker,
        '--result-backend', backend, 'worker', '--concurrency=1'],
        close_fds=True, env=env)
    yield True
    proc.terminate()
    proc.wait()
    Setting().unset(WorkerSettings.BROKER)
    Setting().unset(WorkerSettings.BACKEND)


@pytest.fixture
def smallDocker(boundServer, admin, folder):
    resp = boundServer.request(
        path='/slicer_cli_web/docker_image', user=admin, method='PUT',
        params={
            'name': json.dumps(['girder/slicer_cli_web:small']),
            'folder': folder['_id']
        })
    job = Job().load(id=resp.json['_id'], force=True)
    while job['status'] not in {JobStatus.SUCCESS, JobStatus.CANCELED, JobStatus.ERROR}:
        time.sleep(0.1)
        job = Job().load(id=job['_id'], force=True)
    yield {
        'Example1': Item().findOne({'name': 'Example1'}),
        'Example2': Item().findOne({'name': 'Example2'}),
        'Example3': Item().findOne({'name': 'Example3'}),
    }
    boundServer.request(
        path='/slicer_cli_web/docker_image', user=admin, method='DELETE',
        params={
            'name': json.dumps(['girder/slicer_cli_web:small']),
            'delete_from_local_repo': False,
        })



import os
import pprint
import sys

from slicer_cli_web import CLIArgumentParser


def main(args):
    print('>> parsed arguments')
    pprint.pprint(vars(args))


if __name__ == '__main__':
    if len(sys.argv) == 2:
        if sys.argv[1] == '--json':
            json_spec_file = os.path.splitext(sys.argv[0])[0] + '.json'
            with open(json_spec_file) as f:
                print(f.read())
            sys.exit(0)
        if sys.argv[1] == '--yaml':
            yaml_spec_file = os.path.splitext(sys.argv[0])[0] + '.yaml'
            with open(yaml_spec_file) as f:
                print(f.read())
            sys.exit(0)

    main(CLIArgumentParser().parse_args())



import pytest
from pytest_girder.assertions import assertStatus, assertStatusOk

from slicer_cli_web.config import PluginSettings


@pytest.mark.plugin('slicer_cli_web')
def test_default_task_folder(server, admin, folder):
    # Test the setting
    resp = server.request('/system/setting', method='PUT', params={
        'key': PluginSettings.SLICER_CLI_WEB_TASK_FOLDER,
        'value': 'bad value'
    }, user=admin)
    assertStatus(resp, 400)
    resp = server.request('/system/setting', method='PUT', params={
        'key': PluginSettings.SLICER_CLI_WEB_TASK_FOLDER,
        'value': folder['_id']
    }, user=admin)
    assertStatusOk(resp)

    assert PluginSettings.has_task_folder()
    assert PluginSettings.get_task_folder()['_id'] == folder['_id']



import re
from xml.sax.saxutils import unescape

from girder_worker.docker.io import StreamWriter


class CLIProgressCLIWriter(StreamWriter):
    def __init__(self, job_manager):
        super().__init__()

        self._job_manager = job_manager
        self._buf = b''
        self._state = None
        re_start = re.compile('<filter-start>.*</filter-start>[\\s]*', re.MULTILINE | re.DOTALL)
        re_end = re.compile('<filter-end>.*</filter-end>[\\s]*', re.MULTILINE | re.DOTALL)
        re_time = re.compile('<filter-time>[^<]*</filter-time>[\\s]*', re.MULTILINE)
        re_stage_progress = re.compile(
            '<filter-stage-progress>[^<]*</filter-stage-progress>[\\s]*', re.MULTILINE)

        self._re_progress = re.compile('<filter-progress>([^<]*)</filter-progress>[\\s]*',
                                       re.MULTILINE)
        self._re_name = re.compile('<filter-name>([^<]*)</filter-name>[\\s]*', re.MULTILINE)
        self._re_comment = re.compile('<filter-comment>([^<]*)</filter-comment>[\\s]*',
                                      re.MULTILINE)

        self._re_clean = [re_start, re_end, re_time, re_stage_progress, self._re_progress,
                          self._re_name, self._re_comment]

        self._last_name = 'Unknown'
        self._last_comment = ''

    def forward(self, buf):
        self._job_manager.write(buf)

    def write(self, buf):
        act = self._buf + buf
        self._buf = b''

        if b'</filter-' not in act:
            if b'<filter-' in act:
                # cache for next time
                self._buf = act
                return
            return self.forward(act)

        t = act.decode('utf-8')

        self._update(t)

        for clean in self._re_clean:
            if not t:
                break
            t = clean.sub('', t)
        if t:
            self.forward(t.encode('utf-8'))

    def _update(self, text):
        name = self._re_name.findall(text)
        comment = self._re_comment.findall(text)
        progress = self._re_progress.findall(text)

        if name:
            self._last_name = unescape(name[-1]).strip()
        if comment:
            self._last_comment = unescape(comment[-1]).strip()

        msg = self._last_comment if self._last_comment else self._last_name

        if progress:
            val = float(progress[-1])
            self._job_manager.updateProgress(total=1, current=val, message=msg)

    def close(self):
        # flush rest
        if self._buf:
            self.forward(self._buf)
            self._buf = ''



import io
from xml.etree.ElementTree import Element, ElementTree, SubElement


def _copy(parent, data, *keys, **kwargs):
    for key in keys:
        if key in data:
            SubElement(parent, key).text = str(data[key])

    for key, value in kwargs.items():
        if key in data:
            SubElement(parent, value).text = str(data[key])


def _copy_attr(parent, data, *keys, **kwargs):
    for key in keys:
        if key in data:
            parent.set(key, str(data[key]))

    for key, value in kwargs.items():
        if key in data:
            parent.set(value, str(data[key]))


def _convert_param(parent, param):
    p = SubElement(parent, param['type'])
    if param.get('multiple'):
        p.set('multiple', 'true')

    _copy(p, param, 'label', 'description', 'name', 'index', 'channel')

    _copy_attr(p, param,
               'coordinateSystem', 'fileExtensions', 'defaultPathMatch',
               'defaultNameMatch', 'defaultRelativePath', 'shapes',
               image_type='type', table_type='type', geometry_type='type',
               transform_type='type')

    for key in ['flag', 'longflag']:
        if key not in param:
            continue
        val = param[key]
        flag = SubElement(p, key)
        if isinstance(val, dict):
            _copy_attr(flag, val, 'alias', 'deprecatedalias')
            flag.text = val['value']
        else:
            flag.text = val

    if 'default' in param:
        default_value = param['default']
        d = SubElement(p, 'default')
        if param['type'].endswith('-vector'):
            d.text = ','.join(str(v) for v in default_value)
        else:
            d.text = str(default_value)

    if 'enumeration' in param:
        for value in param['enumeration']:
            SubElement(p, 'element').text = str(value)

    if 'constraints' in param:
        c = SubElement(p, 'constraints')
        _copy(c, param['constraints'], 'minimum', 'maximum', 'step')

    if 'reference' in param:
        ref = param['reference']
        if isinstance(ref, dict):
            ref_e = SubElement(p, 'reference')
            _copy_attr(ref_e, ref, 'role', 'parameter')
            ref_e.text = ref['value']
        else:
            p.set('reference', ref)


def _convert_group(parent, group):
    parameters = SubElement(parent, 'parameters')
    if group.get('advanced'):
        parameters.set('advanced', 'true')
    _copy(parameters, group, 'label', 'description')

    if 'parameters' not in group:
        return

    for param in group['parameters']:
        _convert_param(parameters, param)


def json_to_xml(data):
    root = Element('executable')
    _copy(root, data, 'category', 'title', 'description', 'version', 'license',
          'contributor', 'acknowledgements',
          documentation_url='documentation-url')

    if 'parameter_groups' in data:
        for group in data['parameter_groups']:
            _convert_group(root, group)

    strio = io.StringIO()
    ElementTree(root).write(strio, 'unicode', True)
    return strio.getvalue()



import datetime
import queue
import subprocess
import threading
import time

import yaml
from girder import events, logger
from girder.models.file import File
from girder.models.item import Item
from girder_jobs.constants import JobStatus
from girder_jobs.models.job import Job

from .config import PluginSettings

_workerConfig = None
_commandQueue = None


class CommandQueue(threading.Thread):
    def __init__(self):
        super().__init__()

        self.daemon = True
        self.terminate = False
        self.queue = queue.Queue()
        self.cond = threading.Condition()

    def run(self):
        while not self.terminate:
            try:
                cmd = self.queue.get(block=False)
                logger.info('Running %s' % cmd)
                out = subprocess.check_output(cmd, shell=True)
                logger.info('Ran %s, got %s' % (cmd, out.decode()))
            except queue.Empty:
                with self.cond:
                    self.cond.wait()
            except Exception:
                logger.exception('Exception when running %s' % cmd)

    def add(self, cmd):
        with self.cond:
            self.queue.put(cmd)
            self.cond.notify()

    def stop(self):
        with self.cond:
            self.terminate = True
            self.cond.notify()

    def __del__(self):
        self.stop()


def _loadWorkerConfig():  # noqa
    """
    Load and update the config file.
    """
    global _commandQueue, _workerConfig

    item = PluginSettings.get_worker_config_item()
    if item is None or len(list(Item().childFiles(item, limit=2))) != 1:
        _workerConfig = 'none'
        return
    try:
        file = next(Item().childFiles(item))
        if (isinstance(_workerConfig, dict) and
                file['_id'] == _workerConfig['file']['_id'] and
                file.get('updated', file.get('created')) == _workerConfig['file'].get(
                    'updated', _workerConfig['file'].get('created'))):
            return
        with File().open(file) as fptr:
            config = yaml.safe_load(fptr)
        if not len(config['workers']) >= 1:
            raise Exception('At least one worker must be specified.')
        for entry in config['workers']:
            if 'start' not in entry or 'stop' not in entry:
                raise Exception('All workers must have start and stop values.')
            if 'concurrency' in entry and (
                    int(entry['concurrency']) != entry['concurrency'] or entry['concurrency'] < 1):
                raise Exception('Workers that specify concurrency must be a positive integer.')
        if 'idle-time' in config and not isinstance(config['idle-time'], dict):
            raise Exception('The idle-time entry must be a dictionary.')
    except Exception:
        logger.exception('Failed to load slicer_cli_web worker config item.')
        _workerConfig = 'none'
        return

    stopAll = False
    if not _commandQueue:
        _commandQueue = CommandQueue()
        _commandQueue.start()
    if not isinstance(_workerConfig, dict):
        _workerConfig = {'active': 0, 'start': datetime.datetime.utcnow()}
        # We ask for all workers to stop when we first load a config file
        if config.get('initial-stop') is not False:
            stopAll = True
    _workerConfig['file'] = file
    _workerConfig['workers'] = config['workers']
    _workerConfig['maxConcurrency'] = sum(
        worker.get('concurrency', 1) for worker in config['workers'])
    _workerConfig['idle-time'] = config.get('idle-time', {})
    _workerConfig['idle-time'].setdefault('all', 300)
    _workerConfig['started'] = {
        val for val in _workerConfig.get('started', [])
        if val >= 0 and val < len(config['workers'])}
    logger.info(
        'Loaded slicer_cli_web worker config item with %d worker(s) and %d maxConcurrency',
        len(_workerConfig['workers']), _workerConfig['maxConcurrency'])
    if stopAll:
        for idx in range(len(_workerConfig['workers'])):
            _stopWorker(idx)
    return True


def _startWorker():
    """
    Start any worker.
    """
    for idx in range(len(_workerConfig['workers'])):
        if idx not in _workerConfig['started']:
            cmd = _workerConfig['workers'][idx]['start']
            _workerConfig['started'] |= {idx}
            _commandQueue.add(cmd)
            return


def _stopWorker(idx):
    """
    Stop a worker absed on its index
    """
    cmd = _workerConfig['workers'][idx]['stop']
    _workerConfig['started'] -= {idx}
    _commandQueue.add(cmd)


def _delayStop():
    """
    After a delay, see if we still need to stop.
    """
    if isinstance(_workerConfig, dict):
        _workerConfig.pop('waitThread', None)
        _manageWorkers()


def _stopAllWorkers():
    """
    Check if enough time has passed.  If so, stop all workers.
    """
    if _workerConfig.get('waitThread'):
        return
    delay = 0
    try:
        delay = float(_workerConfig['idle-time']['all'])
    except Exception:
        delay = 300
    delay -= time.time() - _workerConfig['lastChange']
    if delay > 0:
        _workerConfig['waitThread'] = threading.Timer(delay, _delayStop)
        _workerConfig['waitThread'].daemon = True
        _workerConfig['waitThread'].start()
        return
    for worker in list(_workerConfig['started']):
        _stopWorker(worker)


def _manageWorkers(event=None):
    """
    Handle an event that requires us to check on worker managerment.
    """
    if _workerConfig is None:
        _loadWorkerConfig()
    if not isinstance(_workerConfig, dict):
        return
    activeJobs = Job().find({
        'handler': 'celery_handler',
        'status': {'$not': {'$in': [
            JobStatus.SUCCESS, JobStatus.ERROR, JobStatus.CANCELED
        ]}},
        'updated': {'$gte': _workerConfig['start']}
    }).count()

    if activeJobs == _workerConfig['active']:
        if not activeJobs and len(_workerConfig['started']):
            _stopAllWorkers()
        return
    logger.info('Now have %d active job(s) (was %d)' % (activeJobs, _workerConfig['active']))
    _workerConfig['lastChange'] = time.time()
    startedConcurrency = sum(
        _workerConfig['workers'][idx].get('concurrency', 1) for idx in _workerConfig['started'])
    # Start a worker if we have more active jobs than workers and any workers
    # are unstarted.
    if (activeJobs > startedConcurrency and
            min(activeJobs, _workerConfig['maxConcurrency']) > startedConcurrency):
        _startWorker()
    elif not activeJobs:
        _stopAllWorkers()
    _workerConfig['active'] = activeJobs


def _manageWorkersConfig(event):
    """
    Handle an event where the worker configuration may have changed.
    """
    if event.info.get('key') != PluginSettings.SLICER_CLI_WEB_WORKER_CONFIG_ITEM:
        return
    if _loadWorkerConfig():
        _manageWorkers(None)


def _manageWorkersConfigFile(event):
    """
    Handle an event where the worker configuration file may have changed.
    """
    if not isinstance(_workerConfig, dict) or event.info['_id'] != _workerConfig['file']['_id']:
        return
    if _loadWorkerConfig():
        _manageWorkers(None)


def start():
    """
    Add event bindings and start worker management tools.
    """
    events.bind('jobs.cancel', 'slicer_cli_web_worker', _manageWorkers)
    events.bind('jobs.schedule', 'slicer_cli_web_worker', _manageWorkers)
    events.bind('jobs.job.update.after', 'slicer_cli_web_worker', _manageWorkers)
    events.bind('model.job.save.after', 'slicer_cli_web_worker', _manageWorkers)

    events.bind('model.setting.save.after', 'slicer_cli_web_worker', _manageWorkersConfig)
    events.bind('model.file.save.after', 'slicer_cli_web_worker', _manageWorkersConfigFile)
    _manageWorkers(None)



import json
import os

import jinja2
from girder import logger
from girder.api.rest import RestException
from girder.constants import AccessType
from girder.exceptions import FilePathException
from girder.models.file import File
from girder.models.folder import Folder
from girder.models.setting import Setting
from girder.utility.model_importer import ModelImporter

from .cli_utils import (SLICER_TYPE_TO_GIRDER_MODEL_MAP, is_girder_api, is_on_girder,
                        return_parameter_file_name)

OPENAPI_DIRECT_TYPES = {'boolean', 'integer', 'float', 'double', 'string'}
FOLDER_SUFFIX = '_folder'


def _to_file_volume(param, model):
    from girder_worker.docker.transforms.girder import (GirderFileIdToVolume,
                                                        GirderFolderIdToVolume,
                                                        GirderItemIdToVolume)
    from girder_worker.girder_plugin.constants import PluginSettings

    from .girder_worker_plugin.direct_docker_run import DirectGirderFileIdToVolume

    girder_type = SLICER_TYPE_TO_GIRDER_MODEL_MAP[param.typ]

    if girder_type == 'folder':
        return GirderFolderIdToVolume(model['_id'], folder_name=model['name'])
    elif girder_type == 'item':
        return GirderItemIdToVolume(model['_id'])

    if not Setting().get(PluginSettings.DIRECT_PATH):
        return GirderFileIdToVolume(model['_id'], filename=model['name'])

    try:
        path = File().getLocalFilePath(model)
        return DirectGirderFileIdToVolume(model['_id'], direct_file_path=path,
                                          filename=model['name'])
    except FilePathException:
        return GirderFileIdToVolume(model['_id'], filename=model['name'])


def _to_girder_api(param, value):
    from .girder_worker_plugin.direct_docker_run import GirderApiUrl, GirderToken

    if value:
        return value
    if param.name == 'girderApiUrl':
        return GirderApiUrl()
    elif param.name == 'girderToken':
        return GirderToken()
    return value


def _parseParamValue(param, value, user, token):
    if isinstance(value, bytes):
        value = value.decode('utf8')

    param_id = param.identifier()
    if is_on_girder(param):
        girder_type = SLICER_TYPE_TO_GIRDER_MODEL_MAP[param.typ]
        curModel = ModelImporter.model(girder_type)
        loaded = curModel.load(value, level=AccessType.READ, user=user)
        if not loaded:
            raise RestException('Invalid %s id (%s).' % (curModel.name, str(value)))
        return loaded

    try:
        if param.isVector():
            return '%s' % ', '.join(map(str, json.loads(value)))
        elif (param.typ in OPENAPI_DIRECT_TYPES or
                param.typ == 'string-enumeration' or
                param.typ in SLICER_TYPE_TO_GIRDER_MODEL_MAP):
            return str(value)
        else:  # json
            return str(json.loads(value))
    except json.JSONDecodeError:
        msg = 'Error: Parameter value is not in json.dumps format\n' \
              '  Parameter name = %r\n  Parameter type = %r\n' \
              '  Value passed = %r' % (param_id, param.typ, value)
        logger.exception(msg)
        raise RestException(msg)


def _processTemplates(value, param=None, templateParams=None):
    """
    Given a value that could be a string with jinja2-like template values, if
    template parameters are supplied, replace the templated parts of the string
    with the appropriate values.  Chained template references are not resolved
    (e.g., referring to {{parameter_other}} where the other parameter contains
    a template like {{now}} will probably result in the unprocessed template
    value).

    :param value: the object that might contain templates to be replaced.  If
        there are no templates, this is returned unchanged.
    :param param: a cli parameter, optional.  This extends the available
        template keys to include default, description, index, label, and name
        directly from the cli parameter.  'extension' is the first
        fileExtension, if it exists.  'reference' and 'reference_base' are
        mapped to the parameter listed in the cli parameter reference if it is
        populated in the templateParams as 'parameter_<reference>'.
    :param templateParams: a dictionary of template keys and values.  This
        typically contains task, title, image, now, YYYY, MM, DD, hh, mm, ss,
        parameter_<parameter name> and parameter_<parameter name>_base.
    :returns: either the original value if there were no templates, or a string
        with the templates replaced with their values.
    """
    if value == '__default__' and param:
        value = param.default
    if not templateParams:
        return value
    if param:
        templateParams = templateParams.copy()
        for key in {'default', 'description', 'index', 'label', 'name'}:
            if getattr(param, key, None) is not None:
                templateParams[key] = getattr(param, key)
        if getattr(param, 'fileExtensions', None) is not None and len(param.fileExtensions) >= 1:
            templateParams['extension'] = param.fileExtensions[0]
        reference = getattr(param, 'reference', None)
        key = f'parameter_{reference}'
        if key in templateParams:
            templateParams['reference'] = templateParams[key]
            templateParams['reference_base'] = templateParams[f'{key}_base']
    try:
        newvalue = jinja2.Template(str(value)).render(templateParams)
        if newvalue != str(value):
            logger.info('Replaced templated parameter %s with %s.', value, newvalue)
            return newvalue
    except Exception:
        logger.exception('Failed to repalce templated parameter %r', value)
    return value


def _add_optional_input_param(param, args, user, token, templateParams):
    if param.identifier() not in args:
        return []
    value = _parseParamValue(param, args[param.identifier()], user, token)

    container_args = []
    if param.longflag:
        container_args.append(param.longflag)
    elif param.flag:
        container_args.append(param.flag)
    else:
        return []

    if is_on_girder(param):
        # Bindings
        container_args.append(_to_file_volume(param, value))
    elif is_girder_api(param):
        # Bindings
        container_args.append(_to_girder_api(param, value))
    else:
        value = _processTemplates(value, param, templateParams)
        container_args.append(value)
    return container_args


def _add_optional_output_param(param, args, user, result_hooks, reference, templateParams):
    from girder_worker.docker.transforms import VolumePath
    from girder_worker.docker.transforms.girder import GirderUploadVolumePathToFolder

    if (not param.isExternalType() or not is_on_girder(param) or
            param.identifier() not in args or (param.identifier() + FOLDER_SUFFIX) not in args):
        return []
    value = args[param.identifier()]
    value = _processTemplates(value, param, templateParams)
    folder = args[param.identifier() + FOLDER_SUFFIX]

    container_args = []
    if param.longflag:
        container_args.append(param.longflag)
    elif param.flag:
        container_args.append(param.flag)
    else:
        return []

    instance = Folder().load(folder, level=AccessType.WRITE, user=user)
    if not instance:
        raise RestException('Invalid Folder id (%s).' % (str(folder)))

    # Output Binding !!
    path = VolumePath(value)
    container_args.append(path)
    ref = reference.copy()
    ref['identifier'] = param.identifier()
    result_hooks.append(GirderUploadVolumePathToFolder(
        path, folder, must_exist=False,
        upload_kwargs={'reference': json.dumps(ref)}))

    return container_args


def _add_indexed_input_param(param, args, user, token, templateParams=None):
    value = _parseParamValue(param, args[param.identifier()], user, token)

    if is_on_girder(param):
        # Bindings
        return _to_file_volume(param, value), value['name']
    if is_girder_api(param):
        return _to_girder_api(param, value), value['name']
    value = _processTemplates(value, param, templateParams)
    return value, None


def _add_indexed_output_param(param, args, user, result_hooks, reference, templateParams):
    from girder_worker.docker.transforms import VolumePath
    from girder_worker.docker.transforms.girder import GirderUploadVolumePathToFolder

    value = args[param.identifier()]
    value = _processTemplates(value, param, templateParams)
    if param.typ == 'string':
        return value
    folder = args[param.identifier() + FOLDER_SUFFIX]

    instance = Folder().load(folder, level=AccessType.WRITE, user=user)
    if not instance:
        raise RestException('Invalid Folder id (%s).' % (str(folder)))

    # Output Binding
    path = VolumePath(value)
    ref = reference.copy()
    ref['identifier'] = param.identifier()
    result_hooks.append(GirderUploadVolumePathToFolder(
        path, folder, upload_kwargs={'reference': json.dumps(ref)}))
    return path


def _populateTemplateParams(params, user, token, index_params, opt_params, templateParams=None):
    """
    Collect values and keys for processing templates.

    :param params: a dictionary of arguments for the cli.
    :param user: the authenticating user.
    :param token: the authenticating token.
    :param index_params: a list of cli parameters.
    :param opt_params: a list of cli parameters.
    :param templateParams: a dictionary of keys and values to always include in
        the template values, such as 'now', 'task', and 'title'.
    :returns: the collected templateParams.  This will be the supplied template
        parameters plus 'parameter_<name>' and 'parameter_<name>_base' for the
        cli arguments.
    """
    templateParams = templateParams.copy() if templateParams else {}
    for param in index_params + opt_params:
        if param.identifier() in params:
            try:
                value = _parseParamValue(param, params[param.identifier()], user, token)
            except Exception:
                continue
            value = value.get('name') if isinstance(value, dict) else value
            if value:
                templateParams[f'parameter_{param.name}'] = value
                templateParams[f'parameter_{param.name}_base'] = str(value).rsplit('.', 1)[0]
    return templateParams


def _addEnvironmentToTemplateParams(templateParams=None):
    """
    Add the local environment to the template parameters.  Only environment
    variables that start with SLICER_CLI_WEB_ are added, as to do otherwise
    could expose private data.

    :param templateParams: a dictionary of keys and values to always include in
        the template values, such as 'now', 'task', and 'title'.
    :returns: the adjusted templateParams.
    """
    templateParams = templateParams.copy() if templateParams else {}
    for key in os.environ:
        parts = key.split('SLICER_CLI_WEB_', 1)
        if len(parts) == 2:
            templateParams[f'env_{parts[1]}'] = os.environ[key]
    return templateParams


def prepare_task(params, user, token, index_params, opt_params,
                 has_simple_return_file, reference, templateParams=None):
    import uuid

    from girder_worker.docker.transforms import VolumePath
    from girder_worker.docker.transforms.girder import GirderUploadVolumePathToFolder

    uuidVal = str(uuid.uuid4())
    ca = []
    result_hooks = []
    primary_input_name = None

    templateParams = _addEnvironmentToTemplateParams(templateParams)
    templateParams = _populateTemplateParams(
        params, user, token, index_params, opt_params, templateParams)

    # Get primary name and reference
    for param in index_params:
        if param.channel != 'output':
            arg, name = _add_indexed_input_param(param, params, user, token)
            if (name and not primary_input_name and
                    SLICER_TYPE_TO_GIRDER_MODEL_MAP[param.typ] != 'folder'):
                primary_input_name = name
                reference['userId'] = str(user['_id'])
                value = _parseParamValue(param, params[param.identifier()], user, token)
                itemId = value['_id']
                if SLICER_TYPE_TO_GIRDER_MODEL_MAP[param.typ] == 'file':
                    reference['fileId'] = str(value['_id'])
                    itemId = value['itemId']
                reference['itemId'] = str(itemId)
                reference['uuid'] = uuidVal

    # optional params
    for param in opt_params:
        if param.channel == 'output':
            ca.extend(_add_optional_output_param(
                param, params, user, result_hooks, reference, templateParams))
        else:
            ca.extend(_add_optional_input_param(param, params, user, token, templateParams))

    if has_simple_return_file:
        param_id = return_parameter_file_name + FOLDER_SUFFIX
        param_name_id = return_parameter_file_name
        if param_id in params and param_name_id in params:
            value = params[return_parameter_file_name]
            value = _processTemplates(value, templateParams=templateParams)
            folder = params[return_parameter_file_name + FOLDER_SUFFIX]

            instance = Folder().load(folder, level=AccessType.WRITE, user=user)
            if not instance:
                raise RestException('Invalid Folder id (%s).' % (str(folder)))

            ca.append('--returnparameterfile')

            # Output Binding
            path = VolumePath(value)
            ca.append(path)
            ref = reference.copy()
            ref['identifier'] = 'returnparameterfile'
            result_hooks.append(GirderUploadVolumePathToFolder(
                path, folder, upload_kwargs={'reference': json.dumps(ref)}))

    # indexed params
    for param in index_params:
        if param.channel == 'output':
            ca.append(_add_indexed_output_param(
                param, params, user, result_hooks, reference, templateParams))
        else:
            arg, name = _add_indexed_input_param(param, params, user, token, templateParams)
            ca.append(arg)
            if name and not primary_input_name:
                primary_input_name = name

    return ca, result_hooks, primary_input_name



import argparse
import json
import os
import subprocess
import sys
import textwrap as _textwrap

try:
    from girder import logger
except ImportError:
    import logging as logger


class _MultilineHelpFormatter(argparse.HelpFormatter):
    def _fill_text(self, text, width, indent):
        text = self._whitespace_matcher.sub(' ', text).strip()
        paragraphs = text.split('|n')
        multiline_text = ''
        for paragraph in paragraphs:
            formatted_paragraph = '\n' + _textwrap.fill(
                paragraph, width,
                initial_indent=indent,
                subsequent_indent=indent) + '\n'
            multiline_text += formatted_paragraph
        return multiline_text


def _make_print_cli_list_spec_action(cli_list_spec_file):

    with open(cli_list_spec_file) as f:
        str_cli_list_spec = f.read()

    class _PrintCLIListSpecAction(argparse.Action):

        def __init__(self,
                     option_strings,
                     dest=argparse.SUPPRESS,
                     default=argparse.SUPPRESS,
                     help=None):
            super().__init__(
                option_strings=option_strings,
                dest=dest,
                default=default,
                nargs=0,
                help=help)

        def __call__(self, parser, namespace, values, option_string=None):
            print(str_cli_list_spec)
            parser.exit()

    return _PrintCLIListSpecAction


def CLIListEntrypoint(cli_list_spec_file=None, cwd=None):

    if cli_list_spec_file is None:
        cli_list_spec_file = os.path.join(cwd or os.getcwd(), 'slicer_cli_list.json')

    # Parse CLI List spec
    with open(cli_list_spec_file) as f:
        cli_list_spec = json.load(f)

    # create command-line argument parser
    cmdparser = argparse.ArgumentParser(
        formatter_class=_MultilineHelpFormatter
    )

    # add --cli_list
    cmdparser.add_argument(
        '--list_cli',
        action=_make_print_cli_list_spec_action(cli_list_spec_file),
        help='Prints the json file containing the list of CLIs present'
    )

    # add cl-rel-path argument
    cmdparser.add_argument('cli',
                           help='CLI to run',
                           metavar='<cli>',
                           choices=cli_list_spec.keys())

    args = cmdparser.parse_args(sys.argv[1:2])

    args.cli = os.path.normpath(args.cli)

    if cli_list_spec[args.cli]['type'] == 'python':

        script_file = os.path.join(cwd or os.getcwd(), args.cli,
                                   os.path.basename(args.cli) + '.py')

        # python <cli-rel-path>/<cli-name>.py [<args>]
        output_code = subprocess.call([sys.executable, script_file] + sys.argv[2:])

    elif cli_list_spec[args.cli]['type'] == 'cxx':

        script_file = os.path.join(cwd or os.getcwd(), args.cli, os.path.basename(args.cli))

        if os.path.isfile(script_file):

            # ./<cli-rel-path>/<cli-name> [<args>]
            output_code = subprocess.call([script_file] + sys.argv[2:])

        else:

            # assumes parent dir of CLI executable is in ${PATH}
            output_code = subprocess.call([os.path.basename(args.cli)] + sys.argv[2:])

    else:
        logger.exception('CLIs of type %s are not supported',
                         cli_list_spec[args.cli]['type'])
        raise Exception(
            'CLIs of type %s are not supported',
            cli_list_spec[args.cli]['type']
        )

    return output_code


if __name__ == '__main__':
    sys.exit(CLIListEntrypoint())



import ctk_cli

# Allow all parameters of constraints to be optional.
ctk_cli.module.CLIConstraints.OPTIONAL_ELEMENTS = (
    ctk_cli.module.CLIConstraints.OPTIONAL_ELEMENTS +
    ctk_cli.module.CLIConstraints.REQUIRED_ELEMENTS)
ctk_cli.module.CLIConstraints.REQUIRED_ELEMENTS = ()


class CLIParameter(ctk_cli.module.CLIParameter):
    __slots__ = ctk_cli.module.CLIParameter.__slots__ + (
        'datalist',
        'defaultPathMatch',
        'defaultNameMatch',
        'defaultRelativePath',
        'shapes',
    )
    EXTERNAL_TYPES = ctk_cli.module.CLIParameter.EXTERNAL_TYPES.copy()
    EXTERNAL_TYPES.update({
        'item': '.item',  # extension is meaningless
    })
    TYPES = ctk_cli.module.CLIParameter.VALUE_TYPES + tuple(EXTERNAL_TYPES)

    # This is copied from ctk_cli/module.py with a few changes
    @classmethod  # noqa
    def parse(cls, elementTree):  # noqa
        assert ctk_cli.module._tag(elementTree) in cls.TYPES, '%s not in CLIParameter.TYPES' % ctk_cli.module._tag(elementTree)  # noqa

        self = cls()
        self.typ = ctk_cli.module._tag(elementTree)

        if self.typ in ('point', 'region'):
            self._pythonType = float
        else:
            elementType = self.typ
            if elementType.endswith('-vector'):
                elementType = elementType[:-7]
            elif elementType.endswith('-enumeration'):
                elementType = elementType[:-12]
            self._pythonType = self.PYTHON_TYPE_MAPPING.get(elementType, str)

        self.hidden = ctk_cli.module._parseBool(elementTree.get('hidden', 'false'))

        self.constraints = None
        self.multiple = None
        self.elements = None
        self.coordinateSystem = None
        self.fileExtensions = None
        # Added datalist, defaultPathMatch, defaultNameMatch,
        # defaultRelativePath, shapes (slicer_cli_web)
        self.datalist = None
        self.defaultPathMatch = None
        self.defaultNameMatch = None
        self.defaultRelativePath = None
        self.shapes = None
        self.reference = None
        self.subtype = None

        for key, value in elementTree.items():
            if key == 'multiple':
                self.multiple = ctk_cli.module._parseBool(value)
            # Added region (slicer_cli_web)
            elif key == 'coordinateSystem' and self.typ in ('point', 'pointfile', 'region'):
                self.coordinateSystem = value
            elif key == 'fileExtensions':
                self.fileExtensions = [ext.strip() for ext in value.split(',')]
            # Added datalist, defaultPathMatch, defaultNameMatch,
            # default RelativePath (slicer_cli_web)
            elif key == 'datalist':
                self.datalist = value
            elif key == 'defaultPathMatch':
                self.defaultPathMatch = value
            elif key == 'defaultNameMatch':
                self.defaultNameMatch = value
            elif key == 'defaultRelativePath':
                self.defaultRelativePath = value
            elif key == 'shapes':
                self.shapes = value
            # Added file. directory (slicer_cli_web)
            elif key == 'reference' and self.typ in ('image', 'file', 'transform', 'geometry', 'table', 'directory', 'item'):  # noqa
                self.reference = value
                # Suppressed warning (slicer_cli_web)
                # ctk_cli.module.logger.warning("'reference' attribute of %r is not part of the spec yet (CTK issue #623)" % (ctk_cli.module._tag(elementTree), ))  # noqa
            elif key == 'type':
                self.subtype = value
            elif key != 'hidden':
                ctk_cli.module.logger.warning('attribute of %r ignored: %s=%r' % (ctk_cli.module._tag(elementTree), key, value))  # noqa

        elements = []

        childNodes = ctk_cli.module._parseElements(self, elementTree)
        for n in childNodes:
            if ctk_cli.module._tag(n) == 'constraints':
                self.constraints = ctk_cli.module.CLIConstraints.parse(n)
            elif ctk_cli.module._tag(n) == 'element':
                if not n.text:
                    ctk_cli.module.logger.warning('Ignoring empty <element> within <%s>' % (ctk_cli.module._tag(elementTree), ))  # noqa
                else:
                    elements.append(n.text)
            else:
                ctk_cli.module.logger.warning('Element %r within %r not parsed' % (ctk_cli.module._tag(n), ctk_cli.module._tag(elementTree)))  # noqa

        # Added channel and external type checks (slicer_cli_web)
        if (not self.flag and not self.longflag and self.index is None and
                (self.channel != 'output' or self.isExternalType())):
            ctk_cli.module.logger.warning('Parameter %s cannot be passed (missing one of flag, longflag, or index)!' % (  # noqa
                self.identifier(), ))

        if self.flag and not self.flag.startswith('-'):
            self.flag = '-' + self.flag
        if self.longflag and not self.longflag.startswith('-'):
            self.longflag = '--' + self.longflag

        if self.index is not None:
            self.index = int(self.index)

        if self.default:
            try:
                self.default = self.parseValue(self.default)
            except ValueError as e:
                ctk_cli.module.logger.warning('Could not parse default value of <%s> (%s): %s' % (
                    ctk_cli.module._tag(elementTree), self.name, e))

        if self.typ.endswith('-enumeration'):
            try:
                self.elements = list(map(self.parseValue, elements))
            except ValueError as e:
                ctk_cli.module.logger.warning('Problem parsing enumeration element values of <%s> (%s): %s' % (  # noqa
                    ctk_cli.module._tag(elementTree), self.name, e))
            if not elements:
                ctk_cli.module.logger.warning('No <element>s found within <%s>' % (ctk_cli.module._tag(elementTree), ))  # noqa
        else:
            self.elements = None
            if elements:
                ctk_cli.module.logger.warning('Ignoring <element>s within <%s>' % (ctk_cli.module._tag(elementTree), ))  # noqa

        return self


@classmethod  # noqa
def _ctkCliParse(cls, elementTree):  # noqa
    self = cls()

    childNodes = ctk_cli.module._parseElements(self, elementTree, 'parameters')

    self.advanced = ctk_cli.module._parseBool(elementTree.get('advanced', 'false'))

    for pnode in childNodes:
        self.append(CLIParameter.parse(pnode))

    return self


ctk_cli.module.CLIParameters.parse = _ctkCliParse

CLIModule = ctk_cli.CLIModule

__all__ = ['CLIModule']



from os.path import basename

import pytest

from slicer_cli_web.girder_worker_plugin.direct_docker_run import (TEMP_VOLUME_DIRECT_MOUNT_PREFIX,
                                                                   DirectGirderFileIdToVolume, run)


class MockedGirderClient:
    def getFile(self, *args, **kwargs):
        return dict(name='abc')

    def downloadFile(self, file_id, file_path):
        with open(file_path, 'w') as f:
            f.write('dummy')


@pytest.mark.plugin('slicer_cli_web')
def test_direct_docker_run(mocker, server, adminToken, file):
    from girder.models.file import File

    docker_run_mock = mocker.patch(
        'slicer_cli_web.girder_worker_plugin.direct_docker_run._docker_run')
    docker_run_mock.return_value = []

    gc_mock = MockedGirderClient()

    path = File().getLocalFilePath(file)

    run(image='test', container_args=[DirectGirderFileIdToVolume(
        file['_id'], filename=basename(path), direct_file_path=None, gc=gc_mock)])
    docker_run_mock.assert_called_once()
    kwargs = docker_run_mock.call_args[1]
    # image
    assert kwargs['image'] == 'test'
    # container args
    assert len(kwargs['container_args']) == 1
    assert kwargs['container_args'][0].endswith(basename(path))
    # volumes
    assert len(kwargs['volumes']) == 1

    target_path = '%s/%s' % (TEMP_VOLUME_DIRECT_MOUNT_PREFIX, basename(path))

    docker_run_mock.reset_mock()

    run(image='test', container_args=[DirectGirderFileIdToVolume(
        file['_id'], direct_file_path=path, gc=gc_mock)])

    docker_run_mock.assert_called_once()
    kwargs = docker_run_mock.call_args[1]
    # image
    assert kwargs['image'] == 'test'
    # container args
    assert kwargs['container_args'] == [target_path]
    # volumes
    assert len(kwargs['volumes']) == 2



# import json
import pprint

import girder_client
# import large_image
# import numpy
from ctk_cli import CLIArgumentParser


def main(args):
    print('>> parsed arguments')
    pprint.pprint(vars(args))
    gc = girder_client.GirderClient(apiUrl=args.girderApiUrl)
    gc.setToken(args.girderToken)

    annotations = gc.get('annotation', parameters=dict(limit=100, offset=0, itemId=args.imageId))
    pprint.pprint(annotations)


if __name__ == '__main__':
    main(CLIArgumentParser().parse_args())



# requires xmljson

import argparse
import json
import sys
from collections import OrderedDict

import xmljson
import yaml

try:
    from lxml.etree import parse
except ImportError:
    from xml.etree.ElementTree import parse


if __name__ == '__main__':  # noqa
    parser = argparse.ArgumentParser(prog='xmltojson')
    parser.add_argument('src', type=argparse.FileType(), nargs='?',
                        default=sys.stdin, help='Defaults to stdin.')
    parser.add_argument('-o', '--dest', type=argparse.FileType('w'),
                        default=sys.stdout, help='Defaults to stdout.')
    parser.add_argument('-y', '--yaml', action='store_true', default=False,
                        help='Output yaml.')
    args = parser.parse_args()

    data = xmljson.yahoo.data(parse(args.src).getroot())['executable']
    if not args.yaml:
        data['$schema'] = 'slicer_cli_web/models/schema.json'
        if hasattr(data, 'move_to_end'):
            data.move_to_end('$schema', last=False)
    for key, value in list(data.items()):
        if value == {} or value is None:
            data.pop(key)
    paramgroups = data.pop('parameters')
    if not isinstance(paramgroups, list):
        paramgroups = [paramgroups]
    data['parameter_groups'] = []
    for ingroup in paramgroups:
        copylabels = {'label', 'description', 'advanced'}
        group = {k: v for k, v in ingroup.items() if k in copylabels}
        for (k, kcast) in {
            'advanced': lambda x: str(x).lower() == 'true',
        }.items():
            if k in group:
                group[k] = kcast(group[k])
        group['parameters'] = []
        for key, params in ingroup.items():
            if key not in copylabels:
                if not isinstance(params, list):
                    params = [params]
                for param in params:
                    param['type'] = key
                    cast = {
                        'boolean': lambda x: str(x).lower() == 'true',
                        'integer': int,
                        'float': float,
                        'double': float}.get(key.split('-', 1)[0], str)
                    if hasattr(param, 'move_to_end'):
                        param.move_to_end('type', last=False)
                    if key.endswith('-vector') and param.get('default'):
                        param['default'] = [cast(v) for v in param['default'].split(',')]
                    elif param.get('default'):
                        param['default'] = cast(param['default'])
                    if param.get('constraints'):
                        for k, v in param['constraints'].items():
                            param['constraints'][k] = cast(v)
                    group['parameters'].append(param)
                    if key.endswith('-enumeration') and (
                            param.get('element') or param.get('enumeration')):
                        enumer = (param.pop('element') if param.get('element') else
                                  param.pop('enumeration').pop('element'))
                        if not isinstance(enumer, list):
                            enumer = list(enumer)
                        param['enumeration'] = [cast(v) for v in enumer]
                    for (k, kcast) in {
                        'index': int,
                        'multiple': lambda x: str(x).lower() == 'true',
                    }.items():
                        if k in param:
                            param[k] = kcast(param[k])
        data['parameter_groups'].append(group)
    if not args.yaml:
        result = json.dumps(data, indent=2)
    else:
        def represent_dict_order(self, data):
            return self.represent_mapping('tag:yaml.org,2002:map', data.items())

        yaml.add_representer(OrderedDict, represent_dict_order)
        result = yaml.dump(data)
    args.dest.write(result)



import json
from os.path import dirname, join

import yaml
from girder.models.file import File
from jsonschema import validate

from ..cli_utils import as_model
from .json_to_xml import json_to_xml

with open(join(dirname(__file__), 'schema.json')) as f:
    json_schema = json.load(f)


def _parse_xml_desc(item, user, xml):
    meta_data = {
        'xml': xml
    }

    # parse and inject advanced meta data and description
    clim = as_model(xml)
    item['description'] = '**%s**\n\n%s' % (clim.title, clim.description)

    if clim.category:
        meta_data['category'] = clim.category
    if clim.version:
        meta_data['version'] = clim.version
    if clim.license:
        meta_data['license'] = clim.license
    if clim.contributor:
        meta_data['contributor'] = clim.contributor
    if clim.acknowledgements:
        meta_data['acknowledgements'] = clim.acknowledgements

    if clim.documentation_url:
        fileModel = File()
        fileModel.createLinkFile('Documentation', item, 'item',
                                 clim.documentation_url,
                                 user, reuseExisting=True)
    return meta_data


def parse_xml_desc(item, desc, user):
    try:
        return _parse_xml_desc(item, user, desc['xml'])
    except Exception as exc:
        raise Exception('Failed to parse xml (error %s), xml: """%s"""' % (exc, desc['xml']))


def _parse_json_desc(item, user, data):
    validate(data, schema=json_schema)
    xml = json_to_xml(data)
    return _parse_xml_desc(item, user, xml)


def parse_json_desc(item, desc, user):
    try:
        data = json.loads(desc['json'])
    except Exception as exc:
        raise Exception('Failed to load json (error %s), json: """%s"""' % (exc, desc['json']))
    try:
        return _parse_json_desc(item, user, data)
    except Exception as exc:
        raise Exception('Failed to parse json (error %s), json: """%s"""' % (exc, desc['json']))


def parse_yaml_desc(item, desc, user):
    try:
        data = yaml.safe_load(desc['yaml'])
    except Exception as exc:
        raise Exception('Failed to load yaml (error %s), yaml: """%s"""' % (exc, desc['yaml']))
    try:
        return _parse_json_desc(item, user, data)
    except Exception as exc:
        raise Exception('Failed to parse yaml (error %s), yaml: """%s"""' % (exc, desc['yaml']))



from os import R_OK, access
from os.path import abspath, basename, isfile, join

from girder_worker.app import app
from girder_worker.docker.io import FDReadStreamConnector
from girder_worker.docker.tasks import DockerTask, _docker_run
from girder_worker.docker.transforms import BindMountVolume, ContainerStdOut
from girder_worker.docker.transforms.girder import GirderFileIdToVolume
from girder_worker_utils import _walk_obj
from girder_worker_utils.transforms.girder_io import GirderClientTransform

from .cli_progress import CLIProgressCLIWriter


def _get_basename(filename, direct_path):
    if filename:
        return filename
    if not direct_path:
        return None
    return basename(direct_path)


TEMP_VOLUME_DIRECT_MOUNT_PREFIX = '/mnt/girder_direct_worker'


class DirectGirderFileIdToVolume(GirderFileIdToVolume):
    def __init__(self, _id, filename=None, direct_file_path=None, **kwargs):
        superc = super()
        superc.__init__(_id, filename=_get_basename(filename, direct_file_path),
                        **kwargs)
        self._direct_file_path = direct_file_path
        self._direct_container_path = None

    def resolve_direct_file_path(self):
        if not self._direct_file_path:
            return None

        path = abspath(self._direct_file_path)
        if isfile(path) and access(path, R_OK):
            self._direct_container_path = join(TEMP_VOLUME_DIRECT_MOUNT_PREFIX, self._filename)
            return BindMountVolume(path, self._direct_container_path, mode='ro')
        return None

    def transform(self, **kwargs):
        if self._direct_container_path:
            return self._direct_container_path
        return super().transform(**kwargs)


class GirderApiUrl(GirderClientTransform):
    def transform(self, **kwargs):
        return self.gc.urlBase


class GirderToken(GirderClientTransform):
    def transform(self, **kwargs):
        return self.gc.token


def _resolve_direct_file_paths(args, kwargs):
    extra_volumes = []

    def resolve(arg, **kwargs):
        if isinstance(arg, DirectGirderFileIdToVolume):
            path = arg.resolve_direct_file_path()
            if path:
                extra_volumes.append(path)
        return arg
    _walk_obj(args, resolve)
    _walk_obj(kwargs, resolve)

    return extra_volumes


class DirectDockerTask(DockerTask):
    def __call__(self, *args, **kwargs):
        extra_volumes = _resolve_direct_file_paths(args, kwargs)
        if extra_volumes:
            volumes = kwargs.setdefault('volumes', [])
            if isinstance(volumes, list):
                # list mode use
                volumes.extend(extra_volumes)
            else:
                for extra_volume in extra_volumes:
                    volumes.update(extra_volume._repr_json_())

        super().__call__(*args, **kwargs)


def _has_image(image):
    """
    Pulls the specified Docker image onto this worker if not already present
    """
    import docker
    from docker.errors import ImageNotFound

    client = docker.from_env(version='auto')
    try:
        client.images.get(image)
        return True
    except ImageNotFound:
        return False
    finally:
        # close docker client to flush any caches
        client.close()


@app.task(base=DirectDockerTask, bind=True)
def run(task, **kwargs):
    """Wraps docker_run to support direct mount volumnes."""

    pull_image = kwargs.get('pull_image')
    if pull_image == 'if-not-present':
        kwargs['pull_image'] = not _has_image(kwargs['image'])

    if hasattr(task, 'job_manager'):
        stream_connectors = kwargs.setdefault('stream_connectors', [])
        stream_connectors.append(FDReadStreamConnector(
            input=ContainerStdOut(),
            output=CLIProgressCLIWriter(task.job_manager)
        ))

    return _docker_run(task, **kwargs)



import json
import pprint

import large_image
import numpy
from ctk_cli import CLIArgumentParser  # noqa I004

# imported for side effects - prevents spurious warnings
from slicer_cli_web import ctk_cli_adjustment  # noqa


def main(args):
    print('>> parsed arguments')
    pprint.pprint(vars(args))

    ts = large_image.open(args.imageFile)

    tileMeans = []
    tileWeights = []
    # iterate through the tiles at a particular magnification:
    for tile in ts.tileIterator(
            format=large_image.tilesource.TILE_FORMAT_NUMPY,
            scale=dict(magnification=5),
            tile_size=dict(width=2048)):
        # The tile image data is in tile['tile'] and is a numpy
        # multi-dimensional array
        mean = numpy.mean(tile['tile'], axis=(0, 1))
        tileMeans.append(mean)
        tileWeights.append(tile['width'] * tile['height'])
        print('x: %d  y: %d  w: %d  h: %d  mag: %g  color: %s' % (
            tile['x'], tile['y'], tile['width'], tile['height'],
            tile['magnification'], ' '.join(f'{val:g}' for val in mean)))
    mean = numpy.average(tileMeans, axis=0, weights=tileWeights)

    channels = ['red', 'green', 'blue']
    if args.channel in channels:
        average = mean[channels.index(args.channel)]
    else:
        average = float(numpy.average(mean))

    print('Average: %g' % average)
    sampleMetadata = {
        'Average Color': average,
        'Average Color By Band': [float(val) for val in mean],
    }
    open(args.outputItemMetadata, 'w').write(json.dumps(sampleMetadata))

    if args.returnParameterFile:
        with open(args.returnParameterFile, 'w') as f:
            f.write('average = %s\n' % average)


if __name__ == '__main__':
    main(CLIArgumentParser().parse_args())



import argparse
import os
import sys
import time

from progress_helper import ProgressHelper

if __name__ == '__main__':
    if '--json' in sys.argv:
        json_spec_file = os.path.splitext(sys.argv[0])[0] + '.json'
        with open(json_spec_file) as f:
            print(f.read())
        sys.exit(0)

    parser = argparse.ArgumentParser()
    parser.add_argument('--count', dest='count', type=int, default=10)
    parser.add_argument('--sleep', dest='sleep', type=float, default=1)

    args = parser.parse_args()
    with ProgressHelper('Example', 'Sleeping mostly') as p:
        for i in range(args.count):
            print('Sleeping...')
            if i:
                p.message('Sleeping for %g seconds' % ((args.count - i) * args.sleep))
            p.progress(float(i) / args.count)
            time.sleep(args.sleep)
        p.message('Done')
        p.progress(1)



class DockerImageError(Exception):
    def __init__(self, message, image_name='None'):

        self.message = message
        # can be a string or list
        self.imageName = image_name
        Exception.__init__(self, message)

    def __str__(self):
        if isinstance(self.imageName, list):
            return self.message + ' (image names [' + ','.join(self.imageName) + '])'
        elif isinstance(self.imageName, str):
            return self.message + ' (image name: ' + self.imageName + ')'
        else:
            return self.message


class DockerImageNotFoundError(DockerImageError):
    def __init__(self, message, image_name, locations=None):
        super().__init__(message, image_name)
        # list of registries tried(local dockerhub etc )
        self.locations = locations or []



#############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#############################################################################

# These two imports must be in this order for appropriate side effects

# isort: off

from . import ctk_cli_adjustment  # noqa

from ctk_cli import CLIArgumentParser  # noqa

# isort: on


from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as _importlib_version

try:
    __version__ = _importlib_version(__name__)
except PackageNotFoundError:
    # package is not installed
    pass


__license__ = 'Apache 2.0'



import os
import re

import pytest

from slicer_cli_web.models.parser import parse_json_desc, parse_xml_desc, parse_yaml_desc


def read_file(name):
    with open(os.path.join(os.path.dirname(__file__), 'data', name)) as f:
        return f.read()


def read_example_file(example, suffix):
    with open(os.path.join(os.path.dirname(__file__), '..', 'small-docker', example,
                           '%s.%s' % (example, suffix))) as f:
        return f.read()


def assert_string_equal(a, b):
    """
    Assert equal ignoring withspaces

    """
    if a == b:
        return
    assert (a is None) == (b is None)
    if a is None:
        return

    white = re.compile('\\s+')

    a_clean = white.sub(a, '')
    b_clean = white.sub(b, '')

    assert a_clean == b_clean


@pytest.mark.plugin('slicer_cli_web')
class TestParserSimple:
    xml = read_file('parser_simple.xml')
    json = read_file('parser_simple.json')
    yaml = read_file('parser_simple.yaml')

    def verify(self, meta, item):
        from girder.models.item import Item

        assert_string_equal(meta.get('xml'), TestParserSimple.xml)
        assert meta.get('category') == 'A'
        assert meta.get('version') == 'V'
        assert meta.get('license') == 'L'
        assert meta.get('contributor') == 'C'
        assert meta.get('acknowledgements') == 'A'

        assert item.get('description') == '**T**\n\nD'

        files = list(Item().childFiles(item))
        assert len(files) == 1
        link = files[0]
        assert link['name'] == 'Documentation'
        assert link['linkUrl'] == 'https://github.com'

    def test_xml(self, admin, item):
        meta = parse_xml_desc(item, dict(xml=TestParserSimple.xml), admin)
        self.verify(meta, item)

    def test_json(self, admin, item):
        meta = parse_json_desc(item, dict(json=TestParserSimple.json), admin)
        self.verify(meta, item)

    def test_yaml(self, admin, item):
        meta = parse_yaml_desc(item, dict(yaml=TestParserSimple.yaml), admin)
        self.verify(meta, item)


@pytest.mark.plugin('slicer_cli_web')
class TestParserParamsSimple:
    xml = read_file('parser_params_simple.xml')
    json = read_file('parser_params_simple.json')
    yaml = read_file('parser_params_simple.yaml')

    def verify(self, meta, item):
        assert_string_equal(meta.get('xml'), TestParserParamsSimple.xml)

    def test_xml(self, admin, item):
        meta = parse_xml_desc(item, dict(xml=TestParserParamsSimple.xml), admin)
        self.verify(meta, item)

    def test_json(self, admin, item):
        meta = parse_json_desc(item, dict(json=TestParserParamsSimple.json), admin)
        self.verify(meta, item)

    def test_yaml(self, admin, item):
        meta = parse_yaml_desc(item, dict(yaml=TestParserParamsSimple.yaml), admin)
        self.verify(meta, item)


@pytest.mark.plugin('slicer_cli_web')
class TestParserParamsAdvanced:
    xml = read_file('parser_params_advanced.xml')
    json = read_file('parser_params_advanced.json')
    yaml = read_file('parser_params_advanced.yaml')

    def verify(self, meta, item):
        assert_string_equal(meta.get('xml'), TestParserParamsAdvanced.xml)

    def test_xml(self, admin, item):
        meta = parse_xml_desc(item, dict(xml=TestParserParamsAdvanced.xml), admin)
        self.verify(meta, item)

    def test_json(self, admin, item):
        meta = parse_json_desc(item, dict(json=TestParserParamsAdvanced.json), admin)
        self.verify(meta, item)

    def test_yaml(self, admin, item):
        meta = parse_yaml_desc(item, dict(yaml=TestParserParamsAdvanced.yaml), admin)
        self.verify(meta, item)


@pytest.mark.plugin('slicer_cli_web')
class TestParserExample1:
    xml = read_example_file('Example1', 'xml')
    json = read_example_file('Example1', 'json')
    yaml = read_example_file('Example1', 'yaml')

    def verify(self, meta, item):
        assert_string_equal(meta.get('xml'), TestParserExample1.xml)

    def test_xml(self, admin, item):
        meta = parse_xml_desc(item, dict(xml=TestParserExample1.xml), admin)
        self.verify(meta, item)

    def test_json(self, admin, item):
        meta = parse_json_desc(item, dict(json=TestParserExample1.json), admin)
        self.verify(meta, item)

    def test_yaml(self, admin, item):
        meta = parse_yaml_desc(item, dict(yaml=TestParserExample1.yaml), admin)
        self.verify(meta, item)


@pytest.mark.plugin('slicer_cli_web')
class TestParserExample2:
    xml = read_example_file('Example2', 'xml')
    json = read_example_file('Example2', 'json')
    yaml = read_example_file('Example2', 'yaml')

    def verify(self, meta, item):
        assert_string_equal(meta.get('xml'), TestParserExample2.xml)

    def test_xml(self, admin, item):
        meta = parse_xml_desc(item, dict(xml=TestParserExample2.xml), admin)
        self.verify(meta, item)

    def test_json(self, admin, item):
        meta = parse_json_desc(item, dict(json=TestParserExample2.json), admin)
        self.verify(meta, item)

    def test_yaml(self, admin, item):
        meta = parse_yaml_desc(item, dict(yaml=TestParserExample2.yaml), admin)
        self.verify(meta, item)


@pytest.mark.plugin('slicer_cli_web')
class TestParserExample3:
    xml = read_example_file('Example3', 'xml')
    json = read_example_file('Example3', 'json')
    yaml = read_example_file('Example3', 'yaml')

    def verify(self, meta, item):
        assert_string_equal(meta.get('xml'), TestParserExample3.xml)

    def test_xml(self, admin, item):
        meta = parse_xml_desc(item, dict(xml=TestParserExample3.xml), admin)
        self.verify(meta, item)

    def test_json(self, admin, item):
        meta = parse_json_desc(item, dict(json=TestParserExample3.json), admin)
        self.verify(meta, item)

    def test_yaml(self, admin, item):
        meta = parse_yaml_desc(item, dict(yaml=TestParserExample3.yaml), admin)
        self.verify(meta, item)



from .docker_image import CLIItem, DockerImageItem
from .exceptions import DockerImageError, DockerImageNotFoundError

__all__ = ('DockerImageError', 'DockerImageNotFoundError', 'DockerImageItem', 'CLIItem')



# !/usr/bin/env python

###############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

import json
import time

import docker
from girder import logger
from girder.constants import AccessType
from girder.models.folder import Folder
from girder.models.user import User
from girder_jobs.constants import JobStatus
from girder_jobs.models.job import Job

from .models import DockerImageError, DockerImageItem, DockerImageNotFoundError


def deleteImage(job):
    """
    Deletes the docker images specified in the job from the local machine.
    Images are forcefully removed (equivalent to docker rmi -f)
    :param job: The job object specifying the docker images to remove from
    the local machine

    """
    job = Job().updateJob(
        job,
        log='Started to Delete Docker images\n',
        status=JobStatus.RUNNING,
    )
    docker_client = None
    try:
        deleteList = job['kwargs']['deleteList']
        error = False

        try:
            docker_client = docker.from_env(version='auto')

        except docker.errors.DockerException as err:
            logger.exception('Could not create the docker client')
            job = Job().updateJob(
                job,
                log='Failed to create the Docker Client\n' + str(err) + '\n',
                status=JobStatus.ERROR,
            )
            raise DockerImageError('Could not create the docker client')

        for name in deleteList:
            try:
                docker_client.images.remove(name, force=True)

            except Exception as err:
                logger.exception('Failed to remove image')
                job = Job().updateJob(
                    job,
                    log='Failed to remove image \n' + str(err) + '\n',
                )
                error = True
        if error is True:
            job = Job().updateJob(
                job,
                log='Failed to remove some images',
                status=JobStatus.ERROR,
                notify=True,
                progressMessage='Errors deleting some images'
            )
        else:
            job = Job().updateJob(
                job,
                log='Removed all images',
                status=JobStatus.SUCCESS,
                notify=True,
                progressMessage='Removed all images'
            )
    except Exception as err:
        logger.exception('Error with job')
        job = Job().updateJob(
            job,
            log='Error with job \n ' + str(err) + '\n',
            status=JobStatus.ERROR,

        )
    finally:
        if docker_client:
            docker_client.close()


def findLocalImage(client, name):
    """
    Checks if the docker image exist locally
    :param name: The name of the docker image

    :returns id: returns the docker image id
    """
    try:
        image = client.images.get(name)
    except Exception:
        return None
    return image.id


def jobPullAndLoad(job):
    """
    Attempts to cache metadata on images in the pull list and load list.
    Images in the pull list are pulled first, then images in both lists are
    queried for there clis and each cli's xml description. The clis and
    xml data is stored in the girder mongo database
    Event Listeners assume the job is done when the job status
     is ERROR or SUCCESS.
    Event listeners check the jobtype to determine if a job is Dockerimage
    related
    """
    stage = 'initializing'
    try:
        job = Job().updateJob(
            job,
            log='Started to Load Docker images\n',
            status=JobStatus.RUNNING,
        )
        user = User().load(job['userId'], level=AccessType.READ)
        baseFolder = Folder().load(
            job['kwargs']['folder'], user=user, level=AccessType.WRITE, exc=True)

        loadList = job['kwargs']['nameList']

        errorState = False

        notExistSet = set()
        try:
            docker_client = docker.from_env(version='auto')

        except docker.errors.DockerException as err:
            logger.exception('Could not create the docker client')
            job = Job().updateJob(
                job,
                log='Failed to create the Docker Client\n' + str(err) + '\n',
            )
            raise DockerImageError('Could not create the docker client')

        pullList = [
            name for name in loadList
            if not findLocalImage(docker_client, name) or
            str(job['kwargs'].get('pull')).lower() == 'true']
        loadList = [name for name in loadList if name not in pullList]

        try:
            stage = 'pulling'
            pullDockerImage(docker_client, pullList, job)
        except DockerImageNotFoundError as err:
            errorState = True
            notExistSet = set(err.imageName)
            job = Job().updateJob(
                job,
                log='FAILURE: Could not find the following images\n' + '\n'.join(
                    notExistSet) + '\n',
            )
        stage = 'metadata'
        images, loadingError = loadMetadata(job, docker_client, pullList,
                                            loadList, notExistSet)
        for name, cli_dict in images:
            docker_image = docker_client.images.get(name)
            stage = 'parsing'
            DockerImageItem.saveImage(name, cli_dict, docker_image, user, baseFolder)
        if errorState is False and loadingError is False:
            newStatus = JobStatus.SUCCESS
        else:
            newStatus = JobStatus.ERROR
        job = Job().updateJob(
            job,
            log='Finished caching Docker image data\n',
            status=newStatus,
            notify=True,
            progressMessage='Completed caching docker images'
        )
    except Exception as err:
        logger.exception('Error with job with %s', stage)
        job = Job().updateJob(
            job,
            log='Error with job with %s\n %s\n' % (stage, err),
            status=JobStatus.ERROR,
        )


def loadMetadata(job, docker_client, pullList, loadList, notExistSet):
    """
    Attempt to query preexisting images and pulled images for cli data.
    Cli data for each image is stored and returned as a DockerCache Object

    :param Job(): Singleton JobModel used to update job status
    :param job: The current job being executed
    :param docker_client: An instance of the Docker python client
    :param pullList: The list of images that the job attempted to pull
    :param loadList: The list of images to be queried that were already on the
    local machine
    :notExistSet: A subset of the pullList that didnot exist on the Docker
     registry
    or that could not be pulled

    :returns:DockerCache Object containing cli information for each image
    and a boolean indicating whether an error occurred
    """
    # flag to indicate an error occurred
    errorState = False
    images = []
    for name in pullList:
        if name not in notExistSet:
            job = Job().updateJob(
                job,
                log='Image %s was pulled successfully \n' % name,

            )

            try:
                cli_dict = getCliData(name, docker_client, job)
                images.append((name, cli_dict))
                job = Job().updateJob(
                    job,
                    log='Got pulled image %s metadata \n' % name

                )
            except DockerImageError as err:
                job = Job().updateJob(
                    job,
                    log='FAILURE: Error with recently pulled image %s\n%s\n' % (name, err),
                )
                errorState = True

    for name in loadList:
        # create dictionary and load to database
        try:
            cli_dict = getCliData(name, docker_client, job)
            images.append((name, cli_dict))
            job = Job().updateJob(
                job,
                log='Loaded metadata from pre-existing local image %s\n' % name
            )
        except DockerImageError as err:
            job = Job().updateJob(
                job,
                log='FAILURE: Error with recently loading pre-existing image %s\n%s\n' % (
                    name, err),
            )
            errorState = True
    return images, errorState


def getDockerOutput(imgName, command, client):
    """
    Data from each docker image is collected by executing the equivalent of a
    docker run <imgName> <command/args>
    and collecting the output to standard output
    :param imgName: The name of the docker image
    :param command: The commands/ arguments to be passed to the docker image
    :param client: The docker python client
    """
    cont = None
    try:
        cont = client.containers.create(image=imgName, command=command)
        cont.start()
        ret_code = cont.wait()
        if isinstance(ret_code, dict):
            ret_code = ret_code['StatusCode']
        logs = cont.logs(stdout=True, stderr=False, stream=False)
        cont.remove()
    except Exception as err:
        if cont:
            try:
                cont.remove()
            except Exception:
                pass
        logger.exception(
            'Attempt to docker run %s %s failed', imgName, command)
        raise DockerImageError(
            'Attempt to docker run %s %s failed ' % (
                imgName, command) + str(err), imgName)
    if ret_code != 0:
        raise DockerImageError(
            'Attempt to docker run %s %s failed' % (imgName, command), imgName)
    return logs


def getCliData(name, client, job):
    try:
        cli_dict = getDockerOutput(name, '--list_cli', client)
        # contains nested dict
        # {<cliname>:{type:<type>}}
        try:
            if isinstance(cli_dict, bytes):
                cli_dict = cli_dict.decode('utf8')
            cli_dict = json.loads(cli_dict)
        except Exception:
            job = Job().updateJob(
                job,
                log='Failed to parse cli list.  Output of list_cli was\n%r\n' % cli_dict)
            raise
        for key, info in cli_dict.items():
            desc_type = info.get('desc-type', 'xml')
            cli_desc = getDockerOutput(name, [key, f'--{desc_type}'], client)

            if isinstance(cli_desc, bytes):
                cli_desc = cli_desc.decode('utf8')

            cli_dict[key][desc_type] = cli_desc
            job = Job().updateJob(
                job,
                log='Got image %s, cli %s metadata\n' % (name, key),
            )
        return cli_dict
    except Exception as err:
        logger.exception('Error getting %s cli data from image', name)
        raise DockerImageError('Error getting %s cli data from image ' % (name) + str(err))


def pullDockerImage(client, names, job=None):
    """
    Attempt to pull the docker images listed in names. Failure results in a
    DockerImageNotFoundError being raised

    :param client: The docker python client
    :param names: A list of docker images to be pulled from the Dockerhub
    :param job: A job to update with status.
    """
    imgNotExistList = []
    for name in names:
        try:
            logger.info('Pulling %s image', name)
            lastlog = time.time()
            stats = {}
            for line in client.api.pull(name, stream=True, decode=True):
                try:
                    line.update(line.get('progressDetail', {}))
                    if 'id' not in line or ('total' not in line and line['id'] not in stats):
                        continue
                    stats.setdefault(line['id'], line).update(line)
                    if time.time() - lastlog >= 10:
                        total = sum(record['total'] for record in stats.values())
                        downloaded = sum(
                            record['total'] for record in stats.values()
                            if record['status'] != 'Downloading')
                        downloaded += sum(
                            record['current'] for record in stats.values()
                            if record['status'] == 'Downloading')
                        extracted = sum(
                            record['total'] for record in stats.values()
                            if record['status'] == 'Pull complete')
                        extracted += sum(
                            record['current'] for record in stats.values()
                            if record['status'] == 'Extracting')
                        if total:
                            msg = f'Pulling {name} image: '
                            if downloaded < total:
                                val = downloaded
                                msg += 'downloaded '
                            else:
                                val = extracted
                                msg += 'extracted '
                            msg += f'{val}/{total} ({val * 100 / total:4.2f}%)'
                            logger.info(msg)
                            if job:
                                job = Job().updateJob(job, log=msg + '\n')
                        lastlog = time.time()
                except Exception:
                    # Don't fail if the log code has an issue
                    pass
            # some invalid image names will not be pulled but the pull method
            # will not throw an exception so the only way to confirm if a pull
            # succeeded is to attempt a docker inspect on the image
            client.images.get(name)
        except Exception:
            imgNotExistList.append(name)
    if len(imgNotExistList) != 0:
        raise DockerImageNotFoundError('Could not find multiple images ',
                                       image_name=imgNotExistList)



import json
import os
import subprocess
import sys


def processCLI(filename):
    try:
        with open(os.path.join(os.path.dirname(os.path.realpath(__file__)),
                               filename)) as f:
            list_spec = json.load(f)
    except Exception:
        print('Failed to parse %s' % filename)
        return
    if len(sys.argv) >= 2 and sys.argv[1] == '--list_cli':
        print(json.dumps(list_spec, sort_keys=True, indent=2, separators=(',', ': ')))
        return
    if len(sys.argv) < 2 or sys.argv[1][:1] == '-':
        print('%s --list_cli to get a list of available interfaces.' % __file__)
        print('%s <cli> --help for more details.' % __file__)
        return

    cli = os.path.normpath(sys.argv[1])

    cli = list_spec[cli].get('alias', cli)

    if list_spec[cli]['type'] == 'python':
        script_file = os.path.join(cli, os.path.basename(cli) + '.py')
        # python <cli-rel-path>/<cli-name>.py [<args>]
        subprocess.call([sys.executable, script_file] + sys.argv[2:])
    elif list_spec[cli]['type'] == 'cxx':
        script_file = os.path.join('.', cli, os.path.basename(cli))
        # ./<cli-rel-path>/<cli-name> [<args>]
        subprocess.call([script_file] + sys.argv[2:])
    else:
        raise Exception('CLIs of type %s are not supported' % list_spec[cli]['type'])


if __name__ == '__main__':
    processCLI('cli_list.json')



import time

import pytest
import requests
from girder.models.token import Token
from girder_jobs.constants import JobStatus
from girder_jobs.models.job import Job


def waitForBatchJob(job):
    while job['status'] not in {JobStatus.SUCCESS, JobStatus.ERROR, JobStatus.CANCELED}:
        time.sleep(0.1)
        job = Job().load(id=job['_id'], force=True, includeLog=True)
    results = {'job': job}
    if job['status'] == JobStatus.SUCCESS:
        log = ''.join(job['log']) if 'log' in job else ''
        subjobIds = [entry.split(',', 1)[0] for entry in log.split('Scheduling job ')[1:]]
        subjobs = None
        subjobs = [Job().load(id=id, force=True, includeLog=True) for id in subjobIds]
        while any(subjob['status'] not in {
                JobStatus.SUCCESS, JobStatus.ERROR, JobStatus.CANCELED} for subjob in subjobs):
            time.sleep(0.1)
            subjobs = [Job().load(id=id, force=True, includeLog=True) for id in subjobIds]
        results['subjobs'] = subjobs
    return results


def scheduleBatchJob(boundServer, admin, data):
    headers = {
        'Accept': 'application/json',
        'Girder-Token': str(Token().createToken(admin)['_id']),
    }
    req = requests.post(
        'http://127.0.0.1:%d/api/v1/slicer_cli_web/girder_slicer_cli_web_small/Example3/run' %
        boundServer.boundPort, headers=headers, data=data)
    assert req.status_code == 200
    return req


@pytest.mark.usefixtures('girderWorker')
@pytest.mark.plugin('slicer_cli_web')
def testBatchOneParam(boundServer, admin, girderWorker, smallDocker, fileset):
    req = scheduleBatchJob(boundServer, admin, data={
        'file1': '',
        'file1_folder': str(fileset['folder1']['_id']),
        'image1': str(fileset['file1'][0]['_id']),
        'item1': str(fileset['item1'][0]['_id']),
    })
    assert req.status_code == 200
    results = waitForBatchJob(req.json())
    assert results['job']['status'] == JobStatus.SUCCESS
    assert len(results['subjobs']) == len(fileset['file1'])
    assert 'item match' in ''.join(results['subjobs'][0]['log'])
    for subjob in results['subjobs'][1:]:
        assert 'item match' not in ''.join(subjob['log'])


@pytest.mark.usefixtures('girderWorker')
@pytest.mark.plugin('slicer_cli_web')
def testBatchImageParam(boundServer, admin, girderWorker, smallDocker, fileset):
    req = scheduleBatchJob(boundServer, admin, data={
        'file1': str(fileset['file1'][0]['_id']),
        'image1': r'[0-5]',
        'image1_folder': str(fileset['folder1']['_id']),
        'item1': str(fileset['item1'][0]['_id']),
    })
    assert req.status_code == 200
    results = waitForBatchJob(req.json())
    assert results['job']['status'] == JobStatus.SUCCESS
    assert len(results['subjobs']) == 3
    assert 'item match' in ''.join(results['subjobs'][0]['log'])
    for subjob in results['subjobs'][1:]:
        assert 'item match' not in ''.join(subjob['log'])


@pytest.mark.usefixtures('girderWorker')
@pytest.mark.plugin('slicer_cli_web')
def testBatchTwoParams(boundServer, admin, girderWorker, smallDocker, fileset):
    req = scheduleBatchJob(boundServer, admin, data={
        'file1': '',
        'file1_folder': str(fileset['folder1']['_id']),
        'image1': str(fileset['file1'][0]['_id']),
        'item1': '',
        'item1_folder': str(fileset['folder1']['_id']),
    })
    assert req.status_code == 200
    results = waitForBatchJob(req.json())
    assert results['job']['status'] == JobStatus.SUCCESS
    assert len(results['subjobs']) == len(fileset['file1'])
    assert 'item match' in ''.join(results['subjobs'][0]['log'])
    for subjob in results['subjobs'][1:]:
        assert 'item match' not in ''.join(subjob['log'])


@pytest.mark.plugin('slicer_cli_web')
def testBatchMismatchedLists(boundServer, admin, girderWorker, smallDocker, fileset):
    req = scheduleBatchJob(boundServer, admin, data={
        'file1': '[0-3]',
        'file1_folder': str(fileset['folder1']['_id']),
        'image1': '[0-3]',
        'image1_folder': str(fileset['folder1']['_id']),
        'item1': str(fileset['item1'][0]['_id']),
    })
    assert req.status_code == 200
    results = waitForBatchJob(req.json())
    assert results['job']['status'] == JobStatus.ERROR
    assert 'different number of entries on batch inputs' in ''.join(results['job']['log'])


@pytest.mark.usefixtures('girderWorker')
@pytest.mark.plugin('slicer_cli_web')
def testBatchCancel(boundServer, admin, girderWorker, smallDocker, fileset):
    req = scheduleBatchJob(boundServer, admin, data={
        'file1': '',
        'file1_folder': str(fileset['folder1']['_id']),
        'image1': str(fileset['file1'][0]['_id']),
        'item1': str(fileset['item1'][0]['_id']),
    })
    assert req.status_code == 200
    job = Job().load(id=req.json()['_id'], force=True, includeLog=True)
    Job().cancelJob(job)
    results = waitForBatchJob(req.json())
    assert results['job']['status'] == JobStatus.CANCELED



#############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#############################################################################

from girder.api.rest import getCurrentUser
from girder.constants import AccessType
from girder.exceptions import ValidationException
from girder.models.folder import Folder
from girder.models.item import Item
from girder.models.setting import Setting
from girder.settings import SettingDefault
from girder.utility import setting_utilities


# Constants representing the setting keys for this plugin
class PluginSettings:
    SLICER_CLI_WEB_TASK_FOLDER = 'slicer_cli_web.task_folder'
    SLICER_CLI_WEB_WORKER_CONFIG_ITEM = 'slicer_cli_web.worker_config_item'

    @staticmethod
    def has_task_folder():
        return Setting().get(PluginSettings.SLICER_CLI_WEB_TASK_FOLDER) is not None

    @staticmethod
    def get_task_folder():
        folder = Setting().get(PluginSettings.SLICER_CLI_WEB_TASK_FOLDER)
        if not folder:
            return None
        return Folder().load(folder, level=AccessType.READ, user=getCurrentUser())

    @staticmethod
    def has_worker_config_item():
        return Setting().get(PluginSettings.SLICER_CLI_WEB_WORKER_CONFIG_ITEM) is not None

    @staticmethod
    def get_worker_config_item():
        item = Setting().get(PluginSettings.SLICER_CLI_WEB_WORKER_CONFIG_ITEM)
        if not item:
            return None
        return Item().load(item, force=True, exc=False)


@setting_utilities.validator({
    PluginSettings.SLICER_CLI_WEB_TASK_FOLDER
})
def validateFolder(doc):
    if not doc['value']:
        return
    # We should only be able to change settings with admin privilege, so we
    # don't need to check folder access here.
    folder = Folder().load(doc['value'], force=True)
    if not folder:
        raise ValidationException('invalid folder selected')


@setting_utilities.validator({
    PluginSettings.SLICER_CLI_WEB_WORKER_CONFIG_ITEM
})
def validateItem(doc):
    if not doc['value']:
        return
    # We should only be able to change settings with admin privilege, so we
    # don't need to check folder access here.
    item = Item().load(doc['value'], force=True)
    if not item:
        raise ValidationException('invalid folder selected')


# Defaults

# Defaults that have fixed values can just be added to the system defaults
# dictionary.
SettingDefault.defaults.update({
    PluginSettings.SLICER_CLI_WEB_TASK_FOLDER: None,
    PluginSettings.SLICER_CLI_WEB_WORKER_CONFIG_ITEM: None,
})



###############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

from girder.constants import AccessType
from girder.models.file import File
from girder.models.folder import Folder
from girder.models.item import Item

from .parser import parse_json_desc, parse_xml_desc, parse_yaml_desc


def _split(name):
    """
    :param name: image name
    :type name: string
    """
    if ':' in name.split('/')[-1]:
        return name.rsplit(':', 1)
    return name.rsplit('@', 1)


class CLIItem:
    def __init__(self, item):
        self.item = item
        self._id = item['_id']
        self.name = item['name']
        meta = item['meta']
        self.type = meta['type']
        self.image = meta.get('image', 'UNKNOWN')
        self.digest = meta.get('digest', self.image)
        self.xml = meta['xml']
        if isinstance(self.xml, str):
            self.xml = self.xml.encode('utf8')

        self.restBasePath = self.image.replace(':', '_').replace('/', '_').replace('@', '_')
        self.restPath = '%s/%s' % (self.restBasePath, self.name)

    def __str__(self):
        return 'CLIItem %s, image: %s, id: %s' % (self.name, self.image, self._id)

    @staticmethod
    def find(itemId, user):
        itemModel = Item()
        item = itemModel.load(itemId, user=user, level=AccessType.READ)
        if not item:
            return None
        return CLIItem(item)

    @staticmethod
    def _findAllItemImpl(user=None, extraQuery=None):
        itemModel = Item()
        q = {'meta.slicerCLIType': 'task'}

        if extraQuery:
            q.update(extraQuery)

        if user:
            items = itemModel.findWithPermissions(q, user=user, level=AccessType.READ)
        else:
            items = itemModel.find(q)

        return [CLIItem(item) for item in items]

    @staticmethod
    def findAllItems(user=None, baseFolder=None):
        if not baseFolder:
            return CLIItem._findAllItemImpl(user)

        folderModel = Folder()
        graphLookup = {
            'from': 'folder',
            'startWith': '$_id',
            'connectFromField': '_id',
            'connectToField': 'parentId',
            'as': 'children'
        }

        if user and not user['admin']:
            graphLookup['restrictSearchWithMatch'] = folderModel.permissionClauses(user,
                                                                                   AccessType.READ)

        # see https://docs.mongodb.com/manual/reference/operator/aggregation/graphLookup
        r = folderModel.collection.aggregate([
            {
                '$match': dict(_id=baseFolder['_id'])
            },
            {
                '$graphLookup': graphLookup
            },
            {
                '$project': {
                    'leaves': '$children._id',
                }
            }
        ])

        leaves = next(r)['leaves']

        return CLIItem._findAllItemImpl(user, {
            'folderId': {'$in': [baseFolder['_id']] + leaves}
        })


class DockerImageItem:
    def __init__(self, imageFolder, tagFolder, user):
        self.image = imageFolder['name']
        self.tag = tagFolder['name']
        self.name = '%s:%s' % (self.image, self.tag)
        self.imageFolder = imageFolder
        self.tagFolder = tagFolder
        self._id = self.tagFolder['_id']
        self.user = user
        self.name = '%s:%s' % (self.image, self.tag)
        self.digest = tagFolder['meta'].get('digest', self.name)

    def getCLIs(self):
        itemModel = Item()
        q = {
            'meta.slicerCLIType': 'task',
            'folderId': self.tagFolder['_id']
        }
        if self.user:
            items = itemModel.findWithPermissions(q, user=self.user, level=AccessType.READ)
        else:
            items = itemModel.find(q)

        return [CLIItem(item) for item in items]

    @staticmethod
    def find(tagFolderId, user):
        folderModel = Folder()
        tagFolder = folderModel.load(tagFolderId, user=user, level=AccessType.READ)
        if not tagFolder:
            return None
        imageFolder = folderModel.load(tagFolder['parentId'], user=user, level=AccessType.READ)
        return DockerImageItem(imageFolder, tagFolder, user)

    @staticmethod
    def findAllImages(user=None, baseFolder=None):
        folderModel = Folder()
        q = {'meta.slicerCLIType': 'image'}
        if baseFolder:
            q['parentId'] = baseFolder['_id']

        if user:
            imageFolders = folderModel.findWithPermissions(q, user=user, level=AccessType.READ)
        else:
            imageFolders = folderModel.find(q)

        images = []

        for imageFolder in imageFolders:
            qt = {
                'meta.slicerCLIType': 'tag',
                'parentId': imageFolder['_id']
            }
            if user:
                tagFolders = folderModel.findWithPermissions(qt, user=user, level=AccessType.READ)
            else:
                tagFolders = folderModel.find(qt)
            for tagFolder in tagFolders:
                images.append(DockerImageItem(imageFolder, tagFolder, user))
        return images

    @staticmethod
    def removeImages(names, user):
        folderModel = Folder()
        removed = []
        for name in names:
            image, tag = _split(name)
            q = {
                'meta.slicerCLIType': 'image',
                'name': image
            }
            imageFolder = folderModel.findOne(q, user=user, level=AccessType.READ)
            if not imageFolder:
                continue
            qt = {
                'meta.slicerCLIType': 'tag',
                'parentId': imageFolder['_id'],
                'name': tag
            }
            tagFolder = folderModel.findOne(qt, user=user, level=AccessType.WRITE)
            if not tagFolder:
                continue
            folderModel.remove(tagFolder)
            removed.append(name)

            if folderModel.hasAccess(imageFolder, user, AccessType.WRITE) and \
               folderModel.countFolders(imageFolder) == 0:
                # clean also empty image folders
                folderModel.remove(imageFolder)

        return removed

    @staticmethod
    def _create(name, docker_image, user, baseFolder):
        folderModel = Folder()
        fileModel = File()

        imageName, tagName = _split(name)

        image = folderModel.createFolder(baseFolder, imageName,
                                         'Slicer CLI generated docker image folder',
                                         creator=user, reuseExisting=True)
        folderModel.setMetadata(image, dict(slicerCLIType='image'))

        fileModel.createLinkFile('Docker Hub', image, 'folder',
                                 'https://hub.docker.com/r/%s' % imageName,
                                 user, reuseExisting=True)

        tag = folderModel.createFolder(image, tagName,
                                       'Slicer CLI generated docker image tag folder',
                                       creator=user, reuseExisting=True)

        # add docker image labels as meta data
        labels = docker_image.labels.copy()
        if 'description' in labels:
            tag['description'] = labels['description']
            del labels['description']
        labels = {k.replace('.', '_'): v for k, v in labels.items()}
        if 'Author' in docker_image.attrs:
            labels['author'] = docker_image.attrs['Author']

        digest = None
        if docker_image.attrs.get('RepoDigests', []):
            digest = docker_image.attrs['RepoDigests'][0]

        labels['digest'] = digest
        folderModel.setMetadata(tag, labels)

        folderModel.setMetadata(tag, dict(slicerCLIType='tag'))

        return DockerImageItem(image, tag, user)

    @staticmethod
    def _syncItems(image, cli_dict, user):
        folderModel = Folder()
        itemModel = Item()

        children = folderModel.childItems(image.tagFolder, filters={'meta.slicerCLIType': 'task'})
        existingItems = {item['name']: item for item in children}

        for cli, desc in cli_dict.items():
            item = itemModel.createItem(cli, user, image.tagFolder,
                                        'Slicer CLI generated CLI command item',
                                        reuseExisting=True)
            meta_data = dict(slicerCLIType='task', type=desc.get('type', 'Unknown'))

            # copy some things from the image to be independent
            meta_data['image'] = image.name
            meta_data['digest'] = image.digest if image.name != image.digest else None
            if desc.get('docker-params'):
                meta_data['docker-params'] = desc['docker-params']

            desc_type = desc.get('desc-type', 'xml')

            if desc_type == 'xml':
                meta_data.update(parse_xml_desc(item, desc, user))
            elif desc_type == 'yaml':
                meta_data.update(parse_yaml_desc(item, desc, user))
            elif desc_type == 'json':
                meta_data.update(parse_json_desc(item, desc, user))

            itemModel.setMetadata(item, meta_data)

            if cli in existingItems:
                del existingItems[cli]

        # delete superfluous items
        for item in existingItems.values():
            itemModel.remove(item)

    @staticmethod
    def saveImage(name, cli_dict, docker_image, user, baseFolder):
        """
        :param baseFolder
        :type Folder
        """

        image = DockerImageItem._create(name, docker_image, user, baseFolder)
        DockerImageItem._syncItems(image, cli_dict, user)

        return image

    @staticmethod
    def prepare():
        Item().ensureIndex(['meta.slicerCLIType', {'sparse': True}])



import pprint

from slicer_cli_web import CLIArgumentParser


def main(args):
    if args.stringWithOptions == '__datalist__':
        print(
            '<element>First option</element>\n'
            '<element>Second option</element>\n'
            '<element>Third option</element>\n')
        return
    print('>> parsed arguments')
    print('%r' % args)
    pprint.pprint(vars(args), width=1000)
    with open(args.returnParameterFile, 'w') as f:
        f.write('>> parsed arguments\n')
        f.write('%r\n' % args)
    with open(args.arg1, 'w') as f:
        f.write('example\n')


if __name__ == '__main__':
    parser = CLIArgumentParser()
    main(parser.parse_args())



import json

import docker
import pytest
from girder_jobs.constants import JobStatus
from pytest_girder.assertions import assertStatus
from pytest_girder.utils import getResponseBody

from .conftest import splitName


@pytest.mark.plugin('slicer_cli_web')
def testAddNonExistentImage(images):
    # add a bad image
    img_name = 'null/null:null'
    images.assertNoImages()
    images.addImage(img_name, JobStatus.ERROR)
    images.assertNoImages()


@pytest.mark.plugin('slicer_cli_web')
def testDockerAdd(images, server):
    # try to cache a good image to the mongo database
    img_name = 'girder/slicer_cli_web:small'
    images.assertNoImages()
    images.addImage(img_name, JobStatus.SUCCESS)
    images.imageIsLoaded(img_name, True)
    # If checked without a user, we should get an empty list
    resp = server.request(path='/slicer_cli_web/docker_image')
    assertStatus(resp, 200)
    assert json.loads(getResponseBody(resp)) == {}
    images.endpointsExist(img_name, ['Example1', 'Example2', 'Example3'], ['NotAnExample'])
    images.deleteImage(img_name, True)
    images.assertNoImages()


@pytest.mark.plugin('slicer_cli_web')
def testDockerAddBadParam(server, admin, folder):
    # test sending bad parameters to the PUT endpoint
    kwargs = {
        'path': '/slicer_cli_web/docker_image',
        'user': admin,
        'method': 'PUT',
        'params': {'name': json.dumps(6), 'folder': folder['_id']}
    }
    resp = server.request(**kwargs)
    assertStatus(resp, 400)
    assert 'A valid string' in resp.json['message']

    kwargs['params']['name'] = json.dumps({'abc': 'def'})
    resp = server.request(**kwargs)
    assertStatus(resp, 400)
    assert 'A valid string' in resp.json['message']

    kwargs['params']['name'] = json.dumps([6])
    resp = server.request(**kwargs)
    assertStatus(resp, 400)
    assert 'is not a valid string' in resp.json['message']

    kwargs['params']['name'] = '"not json'
    resp = server.request(**kwargs)
    assertStatus(resp, 400)
    assert 'does not have a tag' in resp.json['message']


@pytest.mark.plugin('slicer_cli_web')
def testDockerAddList(images):
    # try to cache a good image to the mongo database
    img_name = 'girder/slicer_cli_web:small'
    images.assertNoImages()
    images.addImage([img_name], JobStatus.SUCCESS)
    images.imageIsLoaded(img_name, True)
    images.deleteImage(img_name, True)
    images.assertNoImages()


@pytest.mark.plugin('slicer_cli_web')
def testDockerAddWithoutVersion(images):
    # all images need a version or hash
    img_name = 'girder/slicer_cli_web'
    images.assertNoImages()
    images.addImage(img_name, None, 400)
    images.assertNoImages()


@pytest.mark.plugin('slicer_cli_web')
def testDockerDelete(images):
    # just delete the metadata in the mongo database
    # don't delete the docker image
    img_name = 'girder/slicer_cli_web:small'
    images.assertNoImages()
    images.addImage(img_name, JobStatus.SUCCESS)
    images.imageIsLoaded(img_name, True)
    images.deleteImage(img_name, True, False)
    images.imageIsLoaded(img_name, exists=False)
    images.assertNoImages()


@pytest.mark.plugin('slicer_cli_web')
def testDockerDeleteFull(images):
    # attempt to delete docker image metadata and the image off the local
    # machine
    img_name = 'girder/slicer_cli_web:small'
    images.assertNoImages()
    images.addImage(img_name, JobStatus.SUCCESS)
    images.imageIsLoaded(img_name, True)
    images.deleteImage(img_name, True, True, JobStatus.SUCCESS)

    try:
        docker_client = docker.from_env(version='auto')
    except Exception as err:
        raise AssertionError('could not create the docker client ' + str(err))

    try:
        docker_client.images.get(img_name)
        raise AssertionError('If the image was deleted then an attempt to get it '
                             'should raise a docker exception')
    except Exception:
        pass

    images.imageIsLoaded(img_name, exists=False)
    images.assertNoImages()


@pytest.mark.plugin('slicer_cli_web')
def testBadImageDelete(images):
    # attempt to delete a non existent image
    img_name = 'null/null:null'
    images.assertNoImages()
    images.deleteImage(img_name, False, )


@pytest.mark.plugin('slicer_cli_web')
def testXmlEndpoint(images, server, admin):
    # loads an image and attempts to run an arbitrary xml endpoint

    img_name = 'girder/slicer_cli_web:small'
    images.addImage(img_name, JobStatus.SUCCESS)
    images.imageIsLoaded(img_name, True)

    name, tag = splitName(img_name)
    data = images.getEndpoint()
    for tag in data.values():
        for cli in tag.values():
            for info in cli.values():
                route = info['xmlspec']
                resp = server.request(
                    path=route,
                    user=admin,
                    isJson=False)
                assertStatus(resp, 200)
                xmlString = getResponseBody(resp)
                # TODO validate with xml schema
                assert xmlString != ''
    images.deleteImage(img_name, True, )


@pytest.mark.plugin('slicer_cli_web')
def testEndpointDeletion(images, server, admin):
    img_name = 'girder/slicer_cli_web:small'
    images.addImage(img_name, JobStatus.SUCCESS)
    images.imageIsLoaded(img_name, True)
    data = images.getEndpoint()
    images.deleteImage(img_name, True)
    name, tag = splitName(img_name)

    for tag in data.values():
        for cli in tag.values():
            for info in cli.values():
                route = info['xmlspec']
                resp = server.request(
                    path=route,
                    user=admin,
                    isJson=False)
                # xml route should have been deleted
                assertStatus(resp, 400)


@pytest.mark.plugin('slicer_cli_web')
def testAddBadImage(images):
    # job should fail gracefully after pulling the image
    img_name = 'library/hello-world:latest'
    images.assertNoImages()
    images.addImage(img_name, JobStatus.ERROR)
    images.assertNoImages()



from unittest import mock

import pytest

from slicer_cli_web.girder_worker_plugin.cli_progress import CLIProgressCLIWriter


@pytest.mark.plugin('slicer_cli_web')
class Test:
    def setup_method(self):
        job_manager = mock.Mock()
        self.write = job_manager.write = mock.MagicMock()
        self.updateProgress = job_manager.updateProgress = mock.MagicMock()
        self.w = CLIProgressCLIWriter(job_manager)

    def test_simple(self):
        self.w.write(b'Test')
        self.write.assert_called_once_with(b'Test')
        self.updateProgress.assert_not_called()

    @pytest.mark.parametrize('tag', [
        b'<filter-start>Dummy</filter-start>',
        b'<filter-end>Dummy</filter-end>',
        b'<filter-time>0.2</filter-time>',
        b'<filter-stage-progress>0.2</filter-stage-progress>',
        b'<filter-name>Dummy</filter-name>',
        b'<filter-comment>Dummy</filter-comment>',
        b"""<filter-start>
            <filter-name>Dummy</filter-name>
            <filter-comment>Dummy</filter-comment>
            </filter-start>""",
        b"""<filter-end>
            <filter-name>Dummy</filter-name>
            <filter-time>0.2</filter-time>
            </filter-end>""",
    ])
    def test_filter(self, tag):
        self.w.write(b'%sTest' % tag)
        self.write.assert_called_once_with(b'Test')
        self.updateProgress.assert_not_called()

    def test_progress(self):
        self.w.write(b"""<filter-start>
        <filter-name>Dummy</filter-name>
        <filter-comment>Comment</filter-comment>
        </filter-start>Test""")
        self.write.assert_called_once_with(b'Test')
        self.updateProgress.assert_not_called()

        self.write.reset_mock()
        self.w.write(b"""<filter-progress>0.1</filter-progress>Test""")
        self.write.assert_called_once_with(b'Test')
        self.updateProgress.assert_called_once_with(total=1, current=0.1, message='Comment')

    def test_progress_only(self):
        self.w.write(b"""<filter-progress>0.1</filter-progress>Test""")
        self.write.assert_called_once_with(b'Test')
        self.updateProgress.assert_called_once_with(total=1, current=0.1, message='Unknown')

    def test_multi_stage(self):
        self.w.write(b"""<filter-progress>0.1""")
        self.write.assert_not_called()
        self.w.write(b"""</filter-progress>Test""")
        self.write.assert_called_once_with(b'Test')
        self.updateProgress.assert_called_once_with(total=1, current=0.1, message='Unknown')

    def test_multi_stage_invalid(self):
        self.w.write(b"""<filter-progress>0.1""")
        self.write.assert_not_called()
        self.w.close()
        self.write.assert_called_once_with(b'<filter-progress>0.1')
        self.updateProgress.assert_not_called()
